# -*- coding: utf-8 -*-
"""
This package provides the qudi logging facility. It facilitates Pythonic logging by joining Qt log
into native Python logging. Also installs a logging handler that emits a Qt Signal to tap into
every time a log event is registered.
Automatically installs all important logging handlers for qudi to work but also allows for
registering (and removal) of additional handlers.
Installs a rotating log file handler to write log messages to disk.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('clear_handlers',
           'get_handler',
           'get_file_handler',
           'get_logger',
           'get_record_table_model',
           'get_signal_handler',
           'get_stderr_handler',
           'init_record_model_handler',
           'init_rotating_file_handler',
           'register_handler',
           'set_log_level',
           'unregister_handler',
           )

import os
import logging
import warnings
from logging.handlers import RotatingFileHandler
from PySide2.QtCore import qInstallMessageHandler

from .handlers import LogSignalHandler, LogTableModelHandler, qt_message_handler


# global variables
# Keep track of all handlers that have been registered to the qudi
_handlers = dict()
# default handlers for qudi root logger
_signal_handler = None
_file_handler = None
_stream_handler = None
_table_model_handler = None
# The qudi root logger for all loggers created with this module API
_qudi_root_logger = None

# Register Qt message handler
qInstallMessageHandler(qt_message_handler)

# initialize logging module
logging.basicConfig(format="%(message)s", level=logging.WARNING)
logging.addLevelName(logging.CRITICAL, 'critical')
logging.addLevelName(logging.ERROR, 'error')
logging.addLevelName(logging.WARNING, 'warning')
logging.addLevelName(logging.INFO, 'info')
logging.addLevelName(logging.DEBUG, 'debug')
logging.addLevelName(logging.NOTSET, 'not set')
logging.captureWarnings(True)

# set level of stream handler which logs to stderr
if len(logging.getLogger().handlers) < 1:
    _stream_handler = logging.StreamHandler()
    logging.getLogger().addHandler(_stream_handler)
else:
    _stream_handler = logging.getLogger().handlers[0]
_stream_handler.setLevel(logging.WARNING)

# Create qudi root logger
_qudi_root_logger = logging.getLogger('qudi')
_qudi_root_logger.setLevel(logging.INFO)
# _qudi_root_logger.propagate = False

# Create and register signal handler in root logger
_signal_handler = LogSignalHandler()
logging.getLogger().addHandler(_signal_handler)


def register_handler(name, handler, silent=False):
    global _handlers
    if name in _handlers:
        if silent:
            unregister_handler(name)
        else:
            raise KeyError(f'Unable to register new logging handler. Handler by name "{name}" '
                           f'already registered.')
    logging.getLogger().addHandler(handler)
    _handlers[name] = handler


def unregister_handler(name, silent=False):
    global _handlers
    handler = _handlers.pop(name, None)
    if handler is None:
        if not silent:
            raise KeyError(
                f'Unable to unregister logging handler. No handler registered by name "{name}".'
            )
    else:
        logging.getLogger().removeHandler(handler)


def clear_handlers():
    global _handlers
    root_logger = logging.getLogger()
    for name in tuple(_handlers):
        root_logger.removeHandler(_handlers.pop(name))


def get_handler(name):
    return _handlers.get(name, None)


def get_logger(name):
    return _qudi_root_logger.getChild(name.split('qudi.', 1)[-1])


def set_log_level(level):
    _signal_handler.setLevel(level)
    _qudi_root_logger.setLevel(level)
    if _table_model_handler is not None:
        _table_model_handler.setLevel(level)
    if _file_handler is not None:
        _file_handler.setLevel(level)


def get_record_table_model():
    return None if _table_model_handler is None else _table_model_handler.table_model


def get_signal_handler():
    return _signal_handler


def get_stderr_handler():
    return _stream_handler


def get_file_handler():
    return _file_handler


def init_record_model_handler(max_records=10000):
    global _table_model_handler

    if _table_model_handler is not None:
        logging.getLogger().removeHandler(_table_model_handler)
        _table_model_handler = None

    _table_model_handler = LogTableModelHandler(level=_qudi_root_logger.level,
                                                max_records=max_records)
    logging.getLogger().addHandler(_table_model_handler)


def init_rotating_file_handler(path='', filename='qudi.log', max_bytes=1024**3, backup_count=5):
    global _file_handler

    # Remove file handler if it has already been registered
    if _file_handler is not None:
        _file_handler.doRollover()
        logging.getLogger().removeHandler(_file_handler)
        _file_handler = None

    filepath = os.path.join(path, filename)
    session_limit = 1000
    session_count = 1
    while session_count <= session_limit:
        try:
            # Start new file if old logfiles exist
            do_rollover = os.path.exists(filepath) and os.stat(filepath).st_size > 0
            _file_handler = RotatingFileHandler(filepath, maxBytes=max_bytes, backupCount=backup_count)
            _file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s',
                                                         datefmt="%Y-%m-%d %H:%M:%S"))
            _file_handler.setLevel(_qudi_root_logger.level)
            if do_rollover:
                _file_handler.doRollover()
        except PermissionError:
            session_count += 1
            if session_count > session_limit:
                warnings.warn(f'Unable to initialize logger rotating file handler. OS denied '
                              f'access to log file or there are more than {session_limit:d} qudi '
                              f'sessions running.')
                return
            split_filename = filename.rsplit('.', 1)
            if len(split_filename) == 2:
                new_filename = f'{split_filename[0]}_session{session_count:d}.{split_filename[1]}'
            else:
                new_filename = f'{filename}_session{session_count:d}'
            filepath = os.path.join(path, new_filename)
        else:
            logging.getLogger().addHandler(_file_handler)
            break



# -*- coding: utf-8 -*-
"""
Decorators and objects used for overloading attributes and interfacing them.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['OverloadedAttribute', 'OverloadProxy']

import weakref
from typing import Any, Callable


class _OverloadedAttributeMapper:
    def __init__(self):
        self._map_dict = dict()
        self._parent = lambda: None

    def add_mapping(self, key: str, obj: Any) -> None:
        self._map_dict[key] = obj

    def remove_mapping(self, key: str) -> None:
        del self._map_dict[key]

    @property
    def parent(self) -> Any:
        return self._parent()

    @parent.setter
    def parent(self, obj: Any) -> None:
        self._parent = weakref.ref(obj)

    def get_mapped(self, key: str) -> Any:
        if key not in self._map_dict:
            raise KeyError(f'No attribute overload found for key "{key}"')
        return self._map_dict[key]

    def __getitem__(self, key):
        mapped_obj = self.get_mapped(key)
        if hasattr(mapped_obj, '__get__'):
            return mapped_obj.__get__(self.parent)
        else:
            return mapped_obj

    def __setitem__(self, key, value):
        mapped_obj = self.get_mapped(key)
        if hasattr(mapped_obj, '__set__'):
            mapped_obj.__set__(self.parent, value)
        else:
            self._map_dict[key] = value

    def __delitem__(self, key):
        mapped_obj = self.get_mapped(key)
        if hasattr(mapped_obj, '__delete__'):
            mapped_obj.__delete__(self.parent)
        else:
            del self._map_dict[key]


class OverloadedAttribute:
    def __init__(self):
        self._attr_mapper = _OverloadedAttributeMapper()

    def overload(self, key: str) -> Callable[[Any], Any]:
        def decorator(attr):
            self._attr_mapper.add_mapping(key, attr)
            return self
        return decorator

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        self._attr_mapper.parent = instance
        return self._attr_mapper

    def __set__(self, instance, value):
        raise AttributeError('can\'t set attribute')

    def __delete__(self, instance):
        raise AttributeError('can\'t delete attribute')

    def setter(self, key: str) -> Callable[[Any], Any]:
        obj = self._attr_mapper.get_mapped(key)

        def decorator(attr):
            self._attr_mapper.add_mapping(key, obj.setter(attr))
            return self

        return decorator

    def getter(self, key: str) -> Callable[[Any], Any]:
        obj = self._attr_mapper.get_mapped(key)

        def decorator(attr):
            self._attr_mapper.add_mapping(key, obj.getter(attr))
            return self

        return decorator

    def deleter(self, key: str) -> Callable[[Any], Any]:
        obj = self._attr_mapper.get_mapped(key)

        def decorator(attr):
            self._attr_mapper.add_mapping(key, obj.deleter(attr))
            return self

        return decorator


class OverloadProxy:
    """Instances of this class serve as proxies for objects containing attributes of type
    OverloadedAttribute. It can be used to hide the overloading mechanism by fixing the overloaded
    attribute access key in a OverloadProxy instance. This allows for interfacing an overloaded
    attribute in the object represented by this proxy by normal "pythonic" means without the
    additional key-mapping lookup usually required by OverloadedAttribute.

    Heavily inspired by this python recipe under PSF License:
    https://code.activestate.com/recipes/496741-object-proxying/
    """

    __slots__ = ['_obj_ref', '_overload_key', '__weakref__']

    def __init__(self, obj: Any, overload_key: str):
        object.__setattr__(self, '_obj_ref', weakref.ref(obj))
        object.__setattr__(self, '_overload_key', overload_key)

    # proxying (special cases)
    def __getattribute__(self, name):
        obj = object.__getattribute__(self, '_obj_ref')()
        attr = getattr(obj, name)
        if isinstance(attr, _OverloadedAttributeMapper):
            return attr[object.__getattribute__(self, '_overload_key')]
        return attr

    def __delattr__(self, name):
        obj = object.__getattribute__(self, '_obj_ref')()
        attr = getattr(obj, name)
        if isinstance(attr, _OverloadedAttributeMapper):
            del attr[object.__getattribute__(self, '_overload_key')]
        else:
            delattr(obj, name)

    def __setattr__(self, name, value):
        obj = object.__getattribute__(self, '_obj_ref')()
        attr = getattr(obj, name)
        if isinstance(attr, _OverloadedAttributeMapper):
            attr[object.__getattribute__(self, '_overload_key')] = value
        else:
            setattr(obj, name, value)

    def __nonzero__(self):
        return bool(object.__getattribute__(self, '_obj_ref')())

    def __str__(self):
        return str(object.__getattribute__(self, '_obj_ref')())

    def __repr__(self):
        return repr(object.__getattribute__(self, '_obj_ref')())

    # factories
    _special_names = (
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', '__contains__',
        '__delitem__', '__delslice__', '__div__', '__divmod__', '__eq__', '__float__',
        '__floordiv__', '__ge__', '__getitem__', '__getslice__', '__gt__', '__hash__', '__hex__',
        '__iadd__', '__iand__', '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__',
        '__imod__', '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', '__long__',
        '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__', '__neg__', '__oct__', '__or__',
        '__pos__', '__pow__', '__radd__', '__rand__', '__rdiv__', '__rdivmod__', '__reduce__',
        '__reduce_ex__', '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__',
        '__rxor__', '__setitem__', '__setslice__', '__sub__', '__truediv__', '__xor__', 'next',
    )

    @classmethod
    def _create_class_proxy(cls, theclass):
        """Creates a proxy for the given class.
        """

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, '_obj_ref')(), name)(*args, **kw)

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name) and not hasattr(cls, name):
                namespace[name] = make_method(name)
        return type(f'{cls.__name__}({theclass.__name__})', (cls,), namespace)

    def __new__(cls, obj, overload_key, *args, **kwargs):
        """Creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are passed to this
        class' __init__, so deriving classes can define an __init__ method of their own.

        note: _class_proxy_cache is unique per class (each deriving class must hold its own cache).
        """
        try:
            cache = cls.__dict__['_class_proxy_cache']
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        return ins


# -*- coding: utf-8 -*-
"""
This file contains a widget to control a ModuleTask and display its state.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['TaskWidget']

import os
from typing import Type, Optional, Dict, Tuple, Any, Iterable
from PySide2 import QtCore, QtWidgets, QtGui

from qudi.util.helpers import is_integer
from qudi.util.paths import get_artwork_dir
from qudi.util.parameters import ParameterWidgetMapper
from qudi.util.widgets.loading_indicator import CircleLoadingIndicator
from qudi.util.widgets.separator_lines import VerticalLine
from qudi.core.scripting.moduletask import ModuleTask


class TestToolButton(QtWidgets.QToolButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setIconSize(event.size())


class TaskWidget(QtWidgets.QWidget):
    """QWidget to control a ModuleTask and display its state.
    """

    sigStartTask = QtCore.Signal(dict)  # parameters
    sigInterruptTask = QtCore.Signal()

    _ParamWidgetsIterable = Iterable[Tuple[QtWidgets.QLabel, QtWidgets.QWidget]]
    _ParamWidgetsDict = Dict[str, Tuple[QtWidgets.QLabel, QtWidgets.QWidget]]

    def __init__(self, *args, task_type: Type[ModuleTask], max_columns: Optional[int] = None,
                 max_rows: Optional[int] = None, **kwargs):
        super().__init__(*args, **kwargs)

        if max_rows is not None and max_columns is not None:
            raise ValueError('Can either set "max_columns" OR "max_rows" but not both.')
        if max_columns is not None and not is_integer(max_columns):
            raise ValueError('"max_columns" must be None or integer value')
        if max_rows is not None and not is_integer(max_rows):
            raise ValueError('"max_rows" must be None or integer value')

        if max_rows is None and max_columns is None:
            max_rows = 8
        elif max_rows is None:
            number_of_widgets = len(task_type.call_parameters())
            max_rows = number_of_widgets // max_columns + number_of_widgets % max_columns

        # Create control button and state label. Arrange them in a sub-layout and connect button.
        # Also add animated busy-indicator
        icon_dir = os.path.join(get_artwork_dir(), 'icons')
        self._play_icon = QtGui.QIcon(os.path.join(icon_dir, 'media-playback-start'))
        self._stop_icon = QtGui.QIcon(os.path.join(icon_dir, 'media-playback-stop'))
        self.state_label = QtWidgets.QLabel('stopped')
        self.state_label.setAlignment(QtCore.Qt.AlignCenter)
        font = self.state_label.font()
        font.setBold(True)
        self.state_label.setFont(font)
        control_width = self.state_label.sizeHint().width() * 2
        self.run_interrupt_button = QtWidgets.QToolButton()
        self.run_interrupt_button.setIcon(self._play_icon)
        self.run_interrupt_button.setToolButtonStyle(QtGui.Qt.ToolButtonIconOnly)
        self.run_interrupt_button.setFixedWidth(control_width)
        self.run_interrupt_button.setFixedHeight(control_width)
        self.run_interrupt_button.setIconSize(self.run_interrupt_button.size())

        self.running_indicator = CircleLoadingIndicator()
        self.running_indicator.setFixedWidth(control_width // 1.5)
        self.running_indicator.setFixedHeight(control_width // 1.5)
        tmp = self.running_indicator.sizePolicy()
        tmp.setRetainSizeWhenHidden(True)
        self.running_indicator.setSizePolicy(tmp)
        ctrl_layout = QtWidgets.QVBoxLayout()
        ctrl_layout.addStretch(1)
        ctrl_layout.addWidget(self.running_indicator, 0, QtCore.Qt.AlignCenter)
        ctrl_layout.addWidget(self.state_label)
        ctrl_layout.addWidget(self.run_interrupt_button, 0, QtCore.Qt.AlignCenter)
        self.running_indicator.hide()

        self.run_interrupt_button.clicked.connect(self._run_interrupt_clicked)

        # Create task parameter editors and put them in a sub-layout
        self.parameter_widgets = self.__create_parameter_editor_widgets(task_type)
        param_layout = self.__layout_parameter_widgets(self.parameter_widgets.values(), max_rows)

        # Add sub-layouts to main layout
        main_layout = QtWidgets.QHBoxLayout()
        main_layout.addLayout(param_layout)
        main_layout.addWidget(VerticalLine())
        main_layout.addLayout(ctrl_layout)
        self.setLayout(main_layout)

        # State flag to indicate current button functionality (start or interrupt)
        self._interrupt_enabled = False

    @staticmethod
    def __create_parameter_editor_widgets(task_type: Type[ModuleTask]) -> _ParamWidgetsDict:
        """Helper function to create editor widgets and labels for each ModuleTask call parameter.
        """
        task_parameters = task_type.call_parameters()
        param_widgets = dict()
        for param_name, param in task_parameters.items():
            editor = ParameterWidgetMapper.widget_for_parameter(param)
            if editor is None:
                editor = QtWidgets.QLabel('Unknown parameter type')
                editor.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            else:
                editor = editor()
                # ToDo: Set default values here
            label = QtWidgets.QLabel(f'{param_name}:')
            label.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
            param_widgets[param_name] = (label, editor)
        return param_widgets

    @staticmethod
    def __layout_parameter_widgets(param_widgets: _ParamWidgetsIterable,
                                   max_rows: int) -> QtWidgets.QGridLayout:
        """Helper function to layout parameter widgets in a QGridLayout."""
        row = 0
        column = 0
        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(1, 1)
        max_height = 0
        for label, editor in param_widgets:
            layout.addWidget(label, row, column)
            layout.addWidget(editor, row, column + 1)
            max_height = max(max_height, editor.sizeHint().height())
            if row + 1 >= max_rows:
                row = 0
                column += 2
                layout.setColumnStretch(column + 1, 1)
            else:
                row += 1
        for ii in range(row):
            layout.setRowMinimumHeight(ii, max_height)
        layout.setRowStretch(row, 1)
        return layout

    @QtCore.Slot()
    def _run_interrupt_clicked(self) -> None:
        """Callback method for button clicks."""
        if self._interrupt_enabled:
            self.sigInterruptTask.emit()
        else:
            self.run_interrupt_button.setEnabled(False)
            self.sigStartTask.emit(self.get_parameters())

    @QtCore.Slot()
    def task_started(self) -> None:
        self._interrupt_enabled = True
        self.run_interrupt_button.setIcon(self._stop_icon)
        self.run_interrupt_button.setEnabled(True)
        self.running_indicator.show()

    @QtCore.Slot(str)
    def task_state_changed(self, new_state: str) -> None:
        """Callback method for ModuleTask state changes."""
        self.state_label.setText(new_state)

    @QtCore.Slot(object, bool)
    def task_finished(self, result: Any, success: bool) -> None:
        """Callback for task finished event."""
        self._interrupt_enabled = False
        self.run_interrupt_button.setIcon(self._play_icon)
        self.run_interrupt_button.setEnabled(True)
        self.running_indicator.hide()
        self.set_task_result(result, success)

    @QtCore.Slot(object, bool)
    def set_task_result(self, result: Any, success: bool) -> None:
        """Updates the task result display."""
        print(result, success)

    def get_parameters(self) -> Dict[str, Any]:
        """Reads parameters from parameter editors and returns them in a dict."""
        parameters = dict()
        for param_name, (_, editor) in self.parameter_widgets.items():
            if isinstance(editor, QtWidgets.QLabel):
                continue
            try:
                parameters[param_name] = editor.value()
            except AttributeError:
                try:
                    parameters[param_name] = editor.isChecked()
                except AttributeError:
                    parameters[param_name] = editor.text()
        return parameters



# -*- coding: utf-8 -*-

"""
This file contains a task class to run with qudi module dependencies as well as various
helper classes to run and manage these tasks.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ModuleTask', 'ModuleTaskStateMachine']


from fysom import Fysom, Canceled
from PySide2 import QtCore
from typing import Mapping, Any, Optional, Callable, Union

from qudi.core.scripting.modulescript import ModuleScript, ModuleScriptInterrupted


class ModuleTaskStateMachine(Fysom, QtCore.QObject):
    """Finite state machine for ModuleTask.
    State diagram for this FSM:

        stopped ---------> starting ---------> running
           ^                  |                   |
           |                  |                   v
           |                  --------------> finishing
           |                                      |
           |                                      |
           -------------------<--------------------
    """
    # do not copy declaration of trigger(self, event, *args, **kwargs), just apply Slot decorator
    trigger = QtCore.Slot(str, result=bool)(Fysom.trigger)

    def __init__(self, callbacks: Mapping[str, Callable[[Any], Union[None, bool]]],
                 parent: Optional[QtCore.QObject] = None):
        # State machine definition
        # the abbreviations for the event list are the following:
        #   name:   event name,
        #   src:    source state,
        #   dst:    destination state
        fsm_cfg = {'initial'  : 'stopped',
                   'events'   : [{'name': 'start', 'src': 'stopped', 'dst': 'starting'},
                                 {'name': 'run', 'src': 'starting', 'dst': 'running'},
                                 {'name': 'finish', 'src': 'running', 'dst': 'finishing'},
                                 {'name': 'terminate', 'src': 'finishing', 'dst': 'stopped'},
                                 {'name': 'skip_run', 'src': 'starting', 'dst': 'finishing'}],
                   'callbacks': dict() if callbacks is None else callbacks}

        # Initialise state machine:
        super().__init__(parent=parent, cfg=fsm_cfg)

    @QtCore.Slot()
    def start(self) -> None:
        super().start()

    @QtCore.Slot()
    def run(self) -> None:
        super().run()

    @QtCore.Slot()
    def finish(self) -> None:
        super().finish()

    @QtCore.Slot()
    def terminate(self) -> None:
        super().terminate()

    @QtCore.Slot()
    def skip_run(self) -> None:
        super().skip_run()


class ModuleTask(ModuleScript):
    """Extends parent ModuleScript class with more functionality like setup and cleanup.
    Includes a finite state machine for better monitoring and control.

    The only part that can be interrupted are the _run() and _setup methods.
    The implementations must occasionally call _check_interrupt() to raise an exception at that
    point if an interrupt is requested. This should happen at points where _cleanup() can
    properly terminate the task afterwards.
    """

    sigStateChanged = QtCore.Signal(str)  # new state name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set up state machine
        fsm_callbacks = {'on_starting': self.__starting_callback,
                         'on_running': self.__running_callback,
                         'on_finishing': self.__finishing_callback,
                         'on_stopped': self.__stopped_callback}
        self._state_machine = ModuleTaskStateMachine(callbacks=fsm_callbacks, parent=self)

    @property
    def state(self):
        return self._state_machine.current

    @QtCore.Slot()
    def run(self) -> None:
        """Kick-off state machine and start task execution.

        DO NOT OVERRIDE IN SUBCLASS!
        """
        if self.running:
            self.log.error(
                f'Unable to run. Task is already running or has been interrupted immediately.'
            )
        else:
            self._state_machine.start()

    # Implement _setup and _cleanup in subclass if needed. By default they will simply do nothing.
    # You MUST in any case implement _run in a subclass (see: ModuleScript._run).
    def _setup(self) -> None:
        """Optional setup procedure to be performed before _run() is called.
        Raising an exception in here will cause the task to directly call _cleanup() and skip the
        _run() call.
        Access (keyword) arguments via self.(kw)args.
        Can NOT be interrupted.

        Implement in subclass.
        """
        pass

    def _cleanup(self) -> None:
        """Optional cleanup procedure to be performed after _setup() and _run() have been called.
        This method will be called even if _setup() or _run() have raised an exception.
        Access (keyword) arguments via self.(kw)args.
        Can NOT be interrupted.

        Implement in subclass.
        """
        pass

    # Callbacks for FSM below. Ignore this part unless you know what you are doing!
    def __starting_callback(self, event: Any) -> None:
        """FSM startup callback. This will call _setup and set status flags accordingly.
        Resets last task result. Handles task interrupts during execution of _setup method.
        """
        self.log.debug(f'Running setup')
        self.sigStateChanged.emit(event.dst)
        self.result = None
        with self._thread_lock:
            self._interrupted = False
            self._success = False
            self._running = True
        skip_run = True
        try:
            self._check_interrupt()
            self._setup()
            self._check_interrupt()
            skip_run = False
        except ModuleScriptInterrupted:
            self.log.info(f'Setup interrupted')
        except:
            self.log.exception('Exception during setup:')
            raise
        finally:
            if skip_run:
                self._state_machine.skip_run()
            else:
                self._state_machine.run()

    def __running_callback(self, event: Any) -> None:
        """FSM callback to execute the mein task _run method. Sets success flag and task result.
        Handles task interrupts during execution of _run method.
        """
        self.log.debug(f'Running main method with\n\targs: {self.args}\n\tkwargs: {self.kwargs}.')
        self.sigStateChanged.emit(event.dst)
        try:
            self._check_interrupt()
            self.result = self._run(*self.args, **self.kwargs)
            with self._thread_lock:
                self._success = True
        except ModuleScriptInterrupted:
            self.log.info(f'Main run method interrupted')
        except:
            self.log.exception('Exception during main run method:')
            raise
        finally:
            self._state_machine.finish()

    def __finishing_callback(self, event: Any) -> None:
        """FSM callback to always call _cleanup method in the end regardless of task success.
        Handles task interrupts during execution of _run method.
        """
        self.log.debug(f'Running cleanup')
        self.sigStateChanged.emit(event.dst)
        try:
            self._cleanup()
        except ModuleScriptInterrupted:
            self.log.info('Cleanup interrupted')
        except:
            self.log.exception('Exception during cleanup:')
            raise
        finally:
            self._state_machine.terminate()

    def __stopped_callback(self, event: Any) -> None:
        """FSM callback to emit a finished Qt signal and reset the running flag after the task has
        been terminated.
        """
        self.log.debug(f'ModuleTask "{self.__class__.__name__}" has been terminated.')
        self.sigStateChanged.emit(event.dst)
        with self._thread_lock:
            self._running = False
        self.sigFinished.emit()



# -*- coding: utf-8 -*-
"""
This file contains the qudi Manager class.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import gc
import sys
import os
import weakref
import inspect
import traceback
import faulthandler
from logging import DEBUG, INFO
from PySide2 import QtCore, QtWidgets

from qudi.core.logger import init_rotating_file_handler, init_record_model_handler, clear_handlers
from qudi.core.logger import get_logger, set_log_level
from qudi.util.paths import get_main_dir, get_default_log_dir
from qudi.util.mutex import Mutex
from qudi.util.colordefs import QudiMatplotlibStyle
from qudi.core.config import Configuration, ValidationError, YAMLError
from qudi.core.watchdog import AppWatchdog
from qudi.core.modulemanager import ModuleManager
from qudi.core.threadmanager import ThreadManager
from qudi.core.gui.gui import Gui
from qudi.core.servers import RemoteModulesServer, QudiNamespaceServer

# Use non-GUI "Agg" backend for matplotlib by default since it is reasonably thread-safe. Otherwise
# you can only plot from main thread and not e.g. in a logic module.
# This causes qudi to not be able to spawn matplotlib GUIs (by calling matplotlib.pyplot.show())
try:
    import matplotlib as _mpl
    _mpl.use('Agg')
except ImportError:
    pass

# Enable the High DPI scaling support of Qt5
os.environ['QT_ENABLE_HIGHDPI_SCALING'] = '1'

if sys.platform == 'win32':
    # Set QT_LOGGING_RULES environment variable to suppress qt.svg related warnings that otherwise
    # spam the log due to some known Qt5 bugs, e.g. https://bugreports.qt.io/browse/QTBUG-52079
    os.environ['QT_LOGGING_RULES'] = 'qt.svg.warning=false'
else:
    # The following will prevent Qt to spam the logs on X11 systems with enough messages
    # to significantly slow the program down. Most of those warnings should have been
    # notice level or lower. This is a known problem since Qt does not fully comply to X11.
    os.environ['QT_LOGGING_RULES'] = '*.debug=false;*.info=false;*.notice=false;*.warning=false'

# Make icons work on non-X11 platforms, import a custom theme
if sys.platform == 'win32':
    try:
        import ctypes
        myappid = 'qudicore-app'  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    except ImportError:
        raise
    except:
        print('SetCurrentProcessExplicitAppUserModelID failed! This is probably not Microsoft '
              'Windows!')

# Set default Qt locale to "C" in order to avoid surprises with number formats and other things
# QtCore.QLocale.setDefault(QtCore.QLocale('en_US'))
QtCore.QLocale.setDefault(QtCore.QLocale.c())


class Qudi(QtCore.QObject):
    """

    """
    _instance = None
    _run_lock = Mutex()
    _quit_lock = Mutex()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None or cls._instance() is None:
            obj = super().__new__(cls, *args, **kwargs)
            cls._instance = weakref.ref(obj)
            return obj
        raise RuntimeError(
            'Only one qudi instance per process possible (Singleton). Please use '
            'Qudi.instance() to get a reference to the already created instance.'
        )

    def __init__(self, no_gui=False, debug=False, log_dir='', config_file=None):
        super().__init__()

        # CLI arguments
        self.no_gui = bool(no_gui)
        self.debug_mode = bool(debug)
        self.log_dir = str(log_dir) if os.path.isdir(log_dir) else get_default_log_dir(
            create_missing=True)

        # Disable pyqtgraph "application exit workarounds" because they cause errors on exit
        try:
            import pyqtgraph
            pyqtgraph.setConfigOption('exitCleanup', False)
        except ImportError:
            pass

        # Enable stack trace output for SIGSEGV, SIGFPE, SIGABRT, SIGBUS and SIGILL signals
        # -> e.g. for segmentation faults
        faulthandler.disable()
        faulthandler.enable(all_threads=True)

        # install logging facility and set logging level
        init_record_model_handler(max_records=10000)
        init_rotating_file_handler(path=self.log_dir)
        set_log_level(DEBUG if self.debug_mode else INFO)

        # Set up logger for qudi main instance
        self.log = get_logger(__class__.__name__)  # will be "qudi.Qudi" in custom logger
        sys.excepthook = self._qudi_excepthook

        # Load configuration from disc if possible
        self.configuration = Configuration()
        try:
            self.configuration.load(config_file, set_default=True)
        except ValueError:
            self.log.info('No qudi configuration file specified. Using empty default config.')
        except (ValidationError, YAMLError):
            self.log.exception('Invalid qudi configuration file specified. '
                               'Falling back to default config.')

        # initialize thread manager and module manager
        self.thread_manager = ThreadManager(parent=self)
        self.module_manager = ModuleManager(qudi_main=self, parent=self)

        # initialize remote modules server if needed
        remote_server_config = self.configuration['remote_modules_server']
        if remote_server_config:
            self.remote_modules_server = RemoteModulesServer(
                parent=self,
                qudi=self,
                name='remote-modules-server',
                host=remote_server_config.get('address', None),
                port=remote_server_config.get('port', None),
                certfile=remote_server_config.get('certfile', None),
                keyfile=remote_server_config.get('certfile', None),
                protocol_config=remote_server_config.get('protocol_config', None),
                ssl_version=remote_server_config.get('ssl_version', None),
                cert_reqs=remote_server_config.get('cert_reqs', None),
                ciphers=remote_server_config.get('ciphers', None),
                force_remote_calls_by_value=self.configuration['force_remote_calls_by_value']
            )
        else:
            self.remote_modules_server = None
        self.local_namespace_server = QudiNamespaceServer(
            parent=self,
            qudi=self,
            name='local-namespace-server',
            port=self.configuration['namespace_server_port'],
            force_remote_calls_by_value=self.configuration['force_remote_calls_by_value']
        )
        self.watchdog = None
        self.gui = None

        self._configured_extension_paths = list()
        self._is_running = False
        self._shutting_down = False

        # Set qudi style for matplotlib
        try:
            import matplotlib.pyplot as plt
            plt.style.use(QudiMatplotlibStyle.style)
        except ImportError:
            pass

    def _qudi_excepthook(self, ex_type, ex_value, ex_traceback):
        """Handler function to be used as sys.excepthook. Should forward all unhandled exceptions
        to logging module.
        """
        # Use default sys.excepthook if exception is exotic/no subclass of Exception.
        if not issubclass(ex_type, Exception):
            sys.__excepthook__(ex_type, ex_value, ex_traceback)
            return

        # Get the most recent traceback
        most_recent_frame = None
        for most_recent_frame, _ in traceback.walk_tb(ex_traceback):
            pass
        # Try to extract the module and class name in which the exception has been raised
        msg = ''
        if most_recent_frame is None:
            logger = self.log
            msg = 'Unhandled qudi exception:'
        else:
            try:
                obj = most_recent_frame.f_locals['self']
                logger = get_logger(f'{obj.__module__}.{obj.__class__.__name__}')
            except (KeyError, AttributeError):
                # Try to extract just the module name in which the exception has been raised
                try:
                    mod = inspect.getmodule(most_recent_frame.f_code)
                    logger = get_logger(mod.__name__)
                except AttributeError:
                    # If no module and class name can be determined, use the application logger
                    logger = self.log
                    msg = 'Unhandled qudi exception:'
        # Log exception with qudi log handler
        logger.error(msg, exc_info=(ex_type, ex_value, ex_traceback))

    @classmethod
    def instance(cls):
        if cls._instance is None:
            return None
        return cls._instance()

    @property
    def is_running(self):
        return self._is_running

    def _remove_extensions_from_path(self):
        # Clean up previously configured expansion paths
        for ext_path in self._configured_extension_paths:
            try:
                sys.path.remove(ext_path)
            except ValueError:
                pass

    def _add_extensions_to_path(self):
        extensions = self.configuration['extension_paths']
        # Add qudi extension paths to sys.path
        insert_index = 1
        for ext_path in reversed(extensions):
            sys.path.insert(insert_index, ext_path)
        self._configured_extension_paths = extensions

    @QtCore.Slot()
    def _configure_qudi(self):
        """
        """
        if self.configuration.file_path is None:
            print('> Applying default configuration...')
            self.log.info('Applying default configuration...')
        else:
            print(f'> Applying configuration from "{self.configuration.file_path}"...')
            self.log.info(f'Applying configuration from "{self.configuration.file_path}"...')

        # Clear all qudi modules
        self.module_manager.clear()

        # Configure extension paths
        self._remove_extensions_from_path()
        self._add_extensions_to_path()

        # Configure qudi modules
        for base in ['hardware', 'logic', 'gui']:
            # Create ManagedModule instance by adding each module to ModuleManager
            for module_name, module_cfg in self.configuration[base].items():
                try:
                    self.module_manager.add_module(name=module_name,
                                                   base=base,
                                                   configuration=module_cfg)
                except:
                    self.module_manager.remove_module(module_name, ignore_missing=True)
                    self.log.exception(f'Unable to create ManagedModule instance for {base} '
                                       f'module "{module_name}"')

        print('> Qudi configuration complete!')
        self.log.info('Qudi configuration complete!')

    def _start_gui(self):
        if self.no_gui:
            return
        self.gui = Gui(qudi_instance=self, stylesheet_path=self.configuration['stylesheet'])
        if not self.configuration['hide_manager_window']:
            self.gui.activate_main_gui()

    def _start_startup_modules(self):
        for module in self.configuration['startup_modules']:
            print(f'> Loading startup module: {module}')
            self.log.info(f'Loading startup module: {module}')
            # Do not crash if a module can not be started
            try:
                self.module_manager.activate_module(module)
            except:
                self.log.exception(f'Unable to activate autostart module "{module}":')

    def run(self):
        """
        """
        with self._run_lock:
            if self._is_running:
                raise RuntimeError('Qudi is already running!')

            # Notify startup
            startup_info = f'Starting qudi{" in debug mode..." if self.debug_mode else "..."}'
            self.log.info(startup_info)
            print(f'> {startup_info}')

            # Get QApplication instance
            app_cls = QtCore.QCoreApplication if self.no_gui else QtWidgets.QApplication
            app = app_cls.instance()
            if app is None:
                app = app_cls(sys.argv)

            # Install app watchdog
            self.watchdog = AppWatchdog(self.interrupt_quit)

            # Start module servers
            if self.remote_modules_server is not None:
                self.remote_modules_server.start()
            self.local_namespace_server.start()

            # Apply configuration to qudi
            self._configure_qudi()

            # Start GUI if needed
            self._start_gui()

            # Start the startup modules defined in the config file
            self._start_startup_modules()

            # Start Qt event loop unless running in interactive mode
            self._is_running = True
            self.log.info('Initialization complete! Starting Qt event loop...')
            print('> Initialization complete! Starting Qt event loop...\n>')
            exit_code = app.exec_()

            self._shutting_down = False
            self._is_running = False
            self.log.info('Shutdown complete! Ciao')
            print('>\n>   Shutdown complete! Ciao.\n>')

            # Exit application
            sys.exit(exit_code)

    def _exit(self, prompt=True, restart=False):
        """Shutdown qudi. Nicely request that all modules shut down if prompt is True.
        Signal restart to parent process (if present) via exitcode 42 if restart is True.
        """
        with self._quit_lock:
            if not self.is_running or self._shutting_down:
                return
            self._shutting_down = True
            if prompt:
                locked_modules = False
                broken_modules = False
                for module in self.module_manager.values():
                    if module.is_busy:
                        locked_modules = True
                    elif module.state == 'BROKEN':
                        broken_modules = True
                    if broken_modules and locked_modules:
                        break

                if self.no_gui:
                    # command line prompt
                    question = '\nSome modules are still locked. ' if locked_modules else '\n'
                    if restart:
                        question += 'Do you really want to restart qudi (y/N)?: '
                    else:
                        question += 'Do you really want to quit qudi (y/N)?: '
                    while True:
                        response = input(question).lower()
                        if response in ('y', 'yes'):
                            break
                        elif response in ('', 'n', 'no'):
                            self._shutting_down = False
                            return
                else:
                    # GUI prompt
                    if restart:
                        if not self.gui.prompt_restart(locked_modules):
                            self._shutting_down = False
                            return
                    else:
                        if not self.gui.prompt_shutdown(locked_modules):
                            self._shutting_down = False
                            return

            QtCore.QCoreApplication.instance().processEvents()
            self.log.info('Qudi shutting down...')
            print('> Qudi shutting down...')
            self.watchdog.terminate()
            self.watchdog = None
            self.log.info('Stopping module server(s)...')
            print('> Stopping module server(s)...')
            if self.remote_modules_server is not None:
                try:
                    self.remote_modules_server.stop()
                except:
                    self.log.exception('Exception during shutdown of remote modules server:')
            try:
                self.local_namespace_server.stop()
            except:
                self.log.exception('Error during shutdown of local namespace server:')
            QtCore.QCoreApplication.instance().processEvents()
            self.log.info('Deactivating modules...')
            print('> Deactivating modules...')
            self.module_manager.stop_all_modules()
            self.module_manager.clear()
            QtCore.QCoreApplication.instance().processEvents()
            if not self.no_gui:
                self.log.info('Closing main GUI...')
                print('> Closing main GUI...')
                self.gui.deactivate_main_gui()
                self.log.info('Closing remaining windows...')
                print('> Closing remaining windows...')
                self.gui.close_windows()
                self.gui.close_system_tray_icon()
                QtCore.QCoreApplication.instance().processEvents()
            self.log.info('Stopping remaining threads...')
            print('> Stopping remaining threads...')
            self.thread_manager.quit_all_threads()
            QtCore.QCoreApplication.instance().processEvents()
            clear_handlers()
            gc.collect()  # Explicit gc call to prevent Qt C++ extensions from using deleted Python objects
            if restart:
                QtCore.QCoreApplication.exit(42)
            else:
                QtCore.QCoreApplication.quit()

    @QtCore.Slot()
    def quit(self):
        self._exit(prompt=False, restart=False)

    @QtCore.Slot()
    def prompt_quit(self):
        self._exit(prompt=True, restart=False)

    @QtCore.Slot()
    def restart(self):
        self._exit(prompt=False, restart=True)

    @QtCore.Slot()
    def prompt_restart(self):
        self._exit(prompt=True, restart=True)

    def interrupt_quit(self):
        if not self._shutting_down:
            self.quit()
            return
        QtCore.QCoreApplication.exit(1)



# -*- coding: utf-8 -*-

"""
This file contains data storage utilities for qudi.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('get_timestamp_filename', 'format_column_headers', 'format_header',
           'metadata_to_str_dict', 'str_dict_to_metadata', 'get_header_from_file',
           'get_info_from_header', 'CsvDataStorage', 'create_dir_for_file', 'DataStorageBase',
           'ImageFormat', 'NpyDataStorage', 'TextDataStorage')

import os
import re
import copy
import numpy as np
import matplotlib.pyplot as plt

from enum import Enum
from datetime import datetime
from abc import ABCMeta, abstractmethod
from matplotlib.backends.backend_pdf import PdfPages
from configparser import ConfigParser
from io import StringIO

from qudi.util.mutex import Mutex
from qudi.util.helpers import is_string_type, is_integer_type, is_float_type, is_complex_type
from qudi.util.helpers import is_string, is_integer, is_float, is_complex, is_number


class ImageFormat(Enum):
    """Image format to use for saving data thumbnails."""

    PNG = '.png'
    PDF = '.pdf'


def get_timestamp_filename(timestamp, nametag=None):
    """
    Returns a qudi standard filename used for saving measurement data to a file,
    not including any file extension.

    Parameters
    ----------
    timestamp : datetime.datetime
        Timestamp used to create the filename from.
    nametag : str, optional
        Additional string to include in the filename.

    Returns
    -------
    str
        Generated filename without file extension.
    """
    # Start of the filename contains the timestamp, i.e. "20210130-1130-59"
    datetime_str = timestamp.strftime('%Y%m%d-%H%M-%S')
    if nametag:
        nametag = nametag.strip()
        # Replace unicode whitespaces with underscores.
        # Consecutive whitespaces are replaced by single underscore.
        nametag = re.sub(r'[\s]+', '_', nametag)
        # ToDo: More character sequence checking needed. Raise exception if bad.
    # Separate nametag and timestamp string with an underscore
    return f'{datetime_str}_{nametag}' if nametag else datetime_str


def format_column_headers(column_headers, delimiter=';;'):
    if isinstance(column_headers, str):
        return column_headers
    if any(not isinstance(header, str) for header in column_headers):
        raise TypeError('column_headers must be iterable of str.')
    return delimiter.join(column_headers)


def metadata_to_str_dict(metadata):
    if metadata:
        return {str(param): repr(value) for param, value in metadata.items()}
    return dict()


def str_dict_to_metadata(str_dict):
    metadata = dict()
    for param, value in str_dict.items():
        try:
            metadata[param] = eval(value)
        except:
            metadata[param] = value
    return metadata


def _is_dtype_class(obj):
    """Helper to check for valid dtypes that can be handled.
    """
    allowed_types = (int,
                     float,
                     complex,
                     str,
                     np.floating,
                     np.integer,
                     np.complexfloating,
                     np.str_,
                     np.string_)
    return type(obj) == type and issubclass(obj, allowed_types)


def _is_dtype_str(obj):
    """Helper to check for valid dtype string.
    """
    return obj in ('int', 'float', 'complex', 'str')


def _value_to_dtype(val):
    """Helper to return the dtype (int, float, complex or str) of a data value.
    """
    if is_string(val):
        return str
    if is_integer(val):
        return int
    if is_float(val):
        return float
    if is_complex(val):
        return complex
    raise TypeError(f'Unknown value dtype encountered: "{type(val)}"')


def _dtype_to_str(obj):
    """Helper to convert dtype class to str representation.
    """
    if _is_dtype_str(obj):
        return obj
    if is_integer_type(obj):
        return 'int'
    if is_float_type(obj):
        return 'float'
    if is_complex_type(obj):
        return 'complex'
    if is_string_type(obj):
        return 'str'
    raise TypeError(f'Invalid dtype encountered: {obj}')


def _str_to_dtype(dtype_str):
    """
    """
    if dtype_str in ('int', 'float', 'complex', 'str'):
        return eval(dtype_str)
    raise ValueError(f'Invalid dtype string encountered: "{dtype_str}".\n'
                     f'Must be one of "int", "float", "complex", "str".')


def _is_1d_array(array):
    try:
        return is_number(array[0]) or is_string(array[0])
    except IndexError:
        raise ValueError('Checking if empty array is 1D is not allowed.')


def format_header(timestamp, number_format=None, metadata=None, notes=None, column_dtypes=None,
                  column_headers=None, comments=None, delimiter=None):
    """
    """
    if comments is None:
        comments = ''
    # Collect all data to include in the header into a config parser
    config = ConfigParser(comment_prefixes=None, delimiters=('=',))

    # write general section
    general_dict = {'timestamp': timestamp.isoformat()}
    if comments:
        general_dict['comments'] = repr(comments)
    if delimiter:
        general_dict['delimiter'] = repr(delimiter)
    if number_format:
        general_dict['number_format'] = number_format
    if column_dtypes:
        if _is_dtype_class(column_dtypes):
            general_dict['column_dtypes'] = _dtype_to_str(column_dtypes)
        elif _is_dtype_str(column_dtypes):
            general_dict['column_dtypes'] = column_dtypes
        else:
            try:
                general_dict['column_dtypes'] = ';;'.join(_dtype_to_str(t) for t in column_dtypes)
            except TypeError:
                raise TypeError(f'Unknown column_dtypes "{column_dtypes}".\nMust either be dtype '
                                f'name str ("int", "float", "complex", "str"), dtype class (int, '
                                f'float, complex, str, numpy.float32, etc.) or sequence of the '
                                f'afore mentioned formats.')
    if column_headers:
        general_dict['column_headers'] = repr(format_column_headers(column_headers))
    if notes:
        general_dict['notes'] = repr(notes)
    config['General'] = general_dict

    # Write user metadata section
    if metadata:
        config['Metadata'] = metadata_to_str_dict(metadata)

    # Write config to string buffer instead of a temporary file
    buffer = StringIO()
    config.write(buffer, space_around_delimiters=False)
    buffer.seek(0)
    header_lines = buffer.read().splitlines()
    buffer.close()

    # Include comment specifiers at the beginning of each line
    # Also add an "end header" marker for easier custom header parsing
    header_lines.append('---- END HEADER ----')
    line_sep = f'\n{comments}'
    return f'{comments}{line_sep.join(header_lines)}\n'


def get_header_from_file(file_path):
    offset = 0
    comments = None
    with open(file_path, 'r') as file:
        for line in file:
            # Determine comments specifier (if there is any)
            if line.endswith('---- END HEADER ----\n'):
                comments = line.rsplit('---- END HEADER ----', 1)[0]
                break
            offset += len(line)
        file.seek(0)
        if comments is None:
            raise RuntimeError(
                'Qudi data file is missing "---- END HEADER ----" marker. File was probably not '
                'created by the same qudi.util.datastorage.<storage class> helper object'
            )
        header_lines = file.read(offset).splitlines()
    line_start = len(comments)
    return '\n'.join(line[line_start:] for line in header_lines), len(header_lines)


def get_info_from_header(header):
    """

    """
    # Parse header sections
    config = ConfigParser(comment_prefixes=None, delimiters=('=',))
    config.read_string(header)

    # extract and convert general section
    general = {'timestamp': config.get('General', 'timestamp', raw=True, fallback=None),
               'comments': config.get('General', 'comments', raw=True, fallback=None),
               'delimiter': config.get('General', 'delimiter', raw=True, fallback=None),
               'number_format': config.get('General', 'number_format', raw=True, fallback=None),
               'column_dtypes': config.get('General', 'column_dtypes', raw=True, fallback=None),
               'column_headers': config.get('General', 'column_headers', raw=True, fallback=None),
               'notes': config.get('General', 'notes', raw=True, fallback=None)}
    if general['timestamp']:
        general['timestamp'] = datetime.fromisoformat(general['timestamp'])
    if general['column_dtypes']:
        dtypes = tuple(_str_to_dtype(t) for t in general['column_dtypes'].split(';;'))
        if len(dtypes) == 1:
            general['column_dtypes'] = dtypes[0]
        elif len(dtypes) > 1:
            general['column_dtypes'] = dtypes
        else:
            general['column_dtypes'] = None
    if general['comments']:
        general['comments'] = eval(general['comments'])
    if general['delimiter']:
        general['delimiter'] = eval(general['delimiter'])
    if general['notes']:
        general['notes'] = eval(general['notes'])
    if general['column_headers']:
        general['column_headers'] = tuple(eval(general['column_headers']).split(';;'))

    # extract metadata
    if config.has_section('Metadata'):
        metadata_str_dict = dict(config.items('Metadata', raw=True))
        metadata = str_dict_to_metadata(metadata_str_dict)
    else:
        metadata = dict()
    return general, metadata


def create_dir_for_file(file_path):
    """Helper method to create the directory (recursively) for a given file path.
    Will NOT raise an error if the directory already exists.

    Parameters
    ----------
    file_path : str
        File path to create the directory for.
    """
    os.makedirs(os.path.dirname(file_path), exist_ok=True)


class DataStorageBase(metaclass=ABCMeta):
    """Base helper class to store/load (measurement)data to/from disk.
    Subclasses handle saving and loading of measurement data (including metadata) for specific file
    formats.
    Metadata is represented as dictionary (key-value pairs).
    It is also possible to set so called "global metadata" using this or any subclass of this class.
    Global metadata is shared and accessible throughout all instances of these storage objects
    within the Python process.

    If the storage type is file based and root_dir is not initialized, each call to save_data must
    provide the full save path information and not just a file name or name tag.
    """
    _global_metadata = dict()
    _global_metadata_lock = Mutex()

    def __init__(self, *, root_dir=None, include_global_metadata=True,
                 image_format=ImageFormat.PNG):
        """
        Parameters
        ----------
        root_dir : str, optional
            Root directory for this storage instance to work in.
        include_global_metadata : bool, optional
            Flag indicating whether to save global metadata.
        image_format : ImageFormat, optional
            Image file format Enum for saving thumbnails.
        """
        if not isinstance(image_format, ImageFormat):
            raise TypeError("image_format must be ImageFormat Enum")

        self.root_dir = root_dir  # ToDo: Maybe some sanity checking for correct path syntax?
        self.include_global_metadata = bool(include_global_metadata)
        self.image_format = image_format

    def save_thumbnail(self, mpl_figure, file_path):
        """Save a matplotlib figure visualizing the saved data in the image format configured.
        It is recommended to use the same file_path as the corresponding data file (if applicable)
        and exclude the file extension (will be added according to image format).

        Parameters
        ----------
        mpl_figure : matplotlib.figure.Figure
            The matplotlib figure object to save as an image.
        file_path : str
            Full file path to use without the file extension.

        Returns
        -------
        str
            Full absolute path of the saved image.
        """
        file_path += self.image_format.value

        if self.image_format is ImageFormat.PDF:
            with PdfPages(file_path) as pdf:
                pdf.savefig(mpl_figure, bbox_inches='tight', pad_inches=0.05)
        elif self.image_format is ImageFormat.PNG:
            mpl_figure.savefig(file_path, bbox_inches='tight', pad_inches=0.05)
        else:
            raise RuntimeError(f'Unknown image format selected: "{self.image_format}"')

        # close matplotlib figure and return
        plt.close(mpl_figure)
        return file_path

    def get_unified_metadata(self, local_metadata=None):
        """Helper method to return a dict containing provided local_metadata as well as global
        metadata depending on include_global_metadata flag.

        Parameters
        ----------
        local_metadata : dict
            Metadata to include in addition to global metadata.
        include_global_metadata : bool, optional
            Flag indicating whether to include global metadata.

        Returns
        -------
        dict
            New dictionary containing local metadata and global metadata.
        """
        metadata = self.get_global_metadata() if self.include_global_metadata else dict()
        if local_metadata is not None:
            metadata.update(local_metadata)
        return metadata

    @abstractmethod
    def save_data(self, data, *, metadata=None, notes=None, nametag=None, timestamp=None, **kwargs):
        """This method must be implemented in a subclass. It should provide the facility to save an
        entire measurement as a whole along with experiment metadata (to include e.g. in the file
        header). The user can either specify an explicit filename or a generic one will be created.
        If optional nametag and/or timestamp is provided, this will be used to create the generic
        filename (only if the filename parameter is omitted).

        Parameters
        ----------
        data : numpy.ndarray
            Data array to be saved (must be 1D or 2D for text files).
        notes : str, optional
            String that is included in the metadata "as-is" without a key.
        metadata : dict, optional
            Named metadata to be saved in the data header / metadata.
        nametag : str, optional
            Nametag to include in the generic filename.
        timestamp : datetime.datetime, optional
            Timestamp to construct a generic filename from.
        filename : str, optional
            Explicit filename to use for saving the data.

        Returns
        -------
        tuple
            Full file path (str), timestamp used (datetime.datetime), saved data shape (tuple).
        """

        pass

    @abstractmethod
    def load_data(self, *args, **kwargs):
        """This method must be implemented in a subclass. It should provide the facility to load a
        saved data set including the metadata/experiment parameters and column headers
        (if possible). Many storage classes can even implement this method as a static method.

        For file-based storage objects, the only parameter should be file_path.

        Parameters
        ----------
        file_path : str
            Path to the file to be loaded.

        Returns
        -------
        np.ndarray
            Data as a numpy array.
        dict
            User metadata.
        dict
            General header data.
        """
        pass

    @classmethod
    def get_global_metadata(cls):
        """Return a copy of the global metadata dict.
        """
        with cls._global_metadata_lock:
            return cls._global_metadata.copy()

    @classmethod
    def add_global_metadata(cls, name, value=None, *, overwrite=False):
        """Set a single global metadata key-value pair or alternatively multiple ones as dict.
        Metadata added this way will persist for all data storage instances in this process until
        being selectively removed by calls to "remove_global_metadata".
        """
        if isinstance(name, str):
            metadata = {name: copy.deepcopy(value)}
        elif isinstance(name, dict):
            if any(not isinstance(key, str) for key in name):
                TypeError('Metadata dict must contain only str type keys.')
            metadata = copy.deepcopy(name)
        else:
            raise TypeError('add_global_metadata expects either a single dict as first argument or '
                            'a str key and a value as first two arguments.')

        with cls._global_metadata_lock:
            if not overwrite:
                duplicate_keys = set(metadata).intersection(cls._global_metadata)
                if duplicate_keys:
                    raise KeyError(f'global metadata keys "{duplicate_keys}" already set while '
                                   f'overwrite flag is False.')
            cls._global_metadata.update(metadata)

    @classmethod
    def remove_global_metadata(cls, names):
        """Remove a global metadata key-value pair by key. Does not raise an error if the key is
        not found.
        """
        if isinstance(names, str):
            names = [names]
        with cls._global_metadata_lock:
            for name in names:
                cls._global_metadata.pop(name, None)


class TextDataStorage(DataStorageBase):
    """Helper class to store (measurement)data on disk as text file.
    Data will always be saved in a tabular format with column headers. Single/Multiple rows are
    appendable.
    """

    # Regular expressions to automatically determine number format
    # __int_regex = re.compile(r'\A[+-]?\d+\Z')
    # __float_regex = re.compile(r'\A[+-]?\d+.\d+([eE][+-]?\d+)?\Z')

    # Default format specifiers for all dtypes
    _default_fmt_for_type = {int: 'd', float: '.15e', complex: 'r', str: 's'}

    def __init__(self, *, root_dir, comments='# ', delimiter='\t', file_extension='.dat',
                 column_formats=None, **kwargs):
        """

        Parameters
        ----------

        root_dir : str
            Root directory for this storage instance to save files into.
        comments : str, optional
            String to put at the beginning of comment and header lines.
        delimiter : str, optional
            Column delimiter used in text files.
        file_extension : str, optional
            File extension to use for text files
        column_formats : str or sequence, optional
            Value format specifier (mini-language) for each column. Single string case will be used for all columns.
        column_headers : str or sequence, optional
            Sequence of strings containing column headers. If a single string is given, write it to file header without
            formatting.
        column_dtypes : str or sequence, optional
            The column dtypes to expect.
        kwargs: optional
            For additional keyword arguments, see DataStorageBase.__init__
        """
        super().__init__(root_dir=root_dir, **kwargs)

        self._file_extension = ''
        self._delimiter = '\t'
        self.file_extension = file_extension
        self.delimiter = delimiter
        self.comments = comments if isinstance(comments, str) else None
        self.column_formats = column_formats

    @property
    def file_extension(self):
        return self._file_extension

    @file_extension.setter
    def file_extension(self, value):
        if (value is not None) and (not isinstance(value, str)):
            raise TypeError('file_extension must be str or None')
        if not value:
            self._file_extension = ''
        elif value.startswith('.'):
            self._file_extension = value
        else:
            self._file_extension = '.' + value

    @property
    def delimiter(self):
        return self._delimiter

    @delimiter.setter
    def delimiter(self, value):
        if not isinstance(value, str) or value == '':
            raise ValueError('delimiter must be non-empty string')
        self._delimiter = value

    def create_header(self, timestamp=None, metadata=None, notes=None, column_headers=None,
                      column_dtypes=None):
        """
        """
        # Gather all metadata (both global and locally provided) into a single dict
        metadata = self.get_unified_metadata(metadata)
        return format_header(timestamp,
                             metadata=metadata,
                             notes=notes,
                             column_headers=column_headers,
                             column_dtypes=column_dtypes,
                             comments=self.comments,
                             delimiter=self.delimiter)

    def new_file(self, *, timestamp=None, metadata=None, notes=None, nametag=None,
                 column_headers=None, column_dtypes=None, filename=None):
        """
        Create a new data file on disk and write header string to it. Will overwrite old files
        silently if they have the same path.

        Parameters
        ----------
        metadata : dict, optional
            Named metadata values to be saved in the data header.
        notes : str, optional
            String that is included in the file header "as-is".
        nametag : str, optional
            Nametag to include in the generic filename.
        timestamp : datetime.datetime, optional
            Timestamp to use. Will create one if missing.
        filename : str, optional
            Custom filename to use (nametag, timestamp, and configured file_extension will not be
            included for file naming).

        Returns
        -------
        tuple
            Full file path (str), timestamp used (datetime.datetime).
        """
        # Create timestamp if missing
        if timestamp is None:
            timestamp = datetime.now()
        # Construct file name if none is given explicitly
        if filename is None:
            filename = get_timestamp_filename(timestamp=timestamp,
                                              nametag=nametag) + self.file_extension
        # Create header
        header = self.create_header(timestamp=timestamp,
                                    metadata=metadata,
                                    notes=notes,
                                    column_headers=column_headers,
                                    column_dtypes=column_dtypes)
        # Determine full file path and create containing directories if needed
        file_path = os.path.join(self.root_dir, filename)
        create_dir_for_file(file_path)
        # Write to file. Overwrite silently.
        with open(file_path, 'w') as file:
            file.write(header)
        return file_path, timestamp

    def append_file(self, data, file_path):
        """Append single or multiple rows to an existing data file.

        Parameters
        ----------
        data : numpy.ndarray
            Data array to be appended.
            For 1D arrays, it represents a single row.
            For 2D arrays, it represents multiple rows.
        file_path : str
            File path to append to.

        Returns
        -------
        tuple
            Number of rows written (int).
            Number of columns written (int).
        """
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f'File to append data to not found: "{file_path}"\n'
                                    f'Create a new file to append to by calling "new_file".')

        # Determine data dimension
        try:
            is_1d = _is_1d_array(data)
        except ValueError:
            # Data array is empty
            return
        # Construct row format specifier
        # Deduce from first data row if no column_formats is configured
        first_row = data if is_1d else data[0]
        number_of_columns = len(first_row)
        if not self.column_formats:
            column_formats = [self._default_fmt_for_type[_value_to_dtype(val)] for val in first_row]
        elif isinstance(self.column_formats, str):
            column_formats = [self.column_formats] * number_of_columns
        elif len(self.column_formats) != number_of_columns:
            raise ValueError(
                'column_formats sequence has not the same length as number of data columns.'
            )
        else:
            column_formats = self.column_formats
        row_fmt_str = self.delimiter.join(f'{{:{fmt}}}' for fmt in column_formats) + '\n'

        # Append data to file
        with open(file_path, 'a') as file:
            # Write data row-by-row
            if is_1d:
                file.write(row_fmt_str.format(*data))
                rows_written = 1
            else:
                rows_written = 0
                for data_row in data:
                    file.write(row_fmt_str.format(*data_row))
                    rows_written += 1
        return rows_written, number_of_columns

    def save_data(self, data, *, timestamp=None, metadata=None, notes=None, nametag=None,
                  column_headers=None, column_dtypes=None, filename=None):
        """See: DataStorageBase.save_data() for more information.

        column_headers : str or list, optional
            Data column header strings or single string.
        """
        # Derive dtypes from first data row if not explicitly given
        if column_dtypes is None:
            first_row = data if _is_1d_array(data) else data[0]
            column_dtypes = [_value_to_dtype(val) for val in first_row]

        # Create new data file (overwrite old one if it exists)
        file_path, timestamp = self.new_file(timestamp=timestamp,
                                             metadata=metadata,
                                             notes=notes,
                                             nametag=nametag,
                                             column_headers=column_headers,
                                             column_dtypes=column_dtypes,
                                             filename=filename)
        # Append data to file
        rows_columns = self.append_file(data, file_path=file_path)
        return file_path, timestamp, rows_columns

    @staticmethod
    def load_data(file_path):
        """ See: DataStorageBase.load_data()

        file_path : str, optional
            Path to file to load data from.
        """
        # Read back metadata
        try:
            header, header_lines = get_header_from_file(file_path)
            general, metadata = get_info_from_header(header)
            # Determine dtype specifier from general header section
            dtype = general['column_dtypes']
            if dtype is not None and not isinstance(dtype, type):
                # If dtypes differ, construct a structured array
                if all(dtype[0] == typ for typ in dtype):
                    dtype = dtype[0]
                elif str in dtype:
                    # handle str type separately since this is (arguably) a bug in numpy.genfromtxt
                    dtype = None
                else:
                    dtype = [(f'f{col:d}', typ) for col, typ in enumerate(dtype)]
            # Load data from file
            data = np.genfromtxt(file_path,
                                 dtype=dtype,
                                 comments=general['comments'],
                                 delimiter=general['delimiter'],
                                 skip_header=header_lines + 1)
        except UnicodeError as err:
            raise ValueError(f'Loading data from file "{file_path}" failed. The file you are '
                             f'trying to load is most likely no unicode textfile.') from err
        return data, metadata, general


class CsvDataStorage(TextDataStorage):
    """Helper class to store (measurement)data on disk as CSV file.
    This is a specialized sub-class of TextDataStorage that uses hard-coded commas as delimiter and
    includes column headers uncommented in the first row of data. This is the standard format for
    importing a table into e.g. MS Excel.
    """

    def __init__(self, *, file_extension='.csv', **kwargs):
        """See: qudi.util.datastorage.TextDataStorage
        """
        kwargs['delimiter'] = ','
        super().__init__(file_extension=file_extension, **kwargs)

    @property
    def delimiter(self):
        return ','

    @delimiter.setter
    def delimiter(self, value):
        if value != ',':
            self._delimiter = ','
            raise UserWarning('CsvDataStorage only accepts "," as delimiter')

    def create_header(self, timestamp=None, metadata=None, notes=None, column_headers=None,
                      column_dtypes=None):
        """Include column_headers without line comment specifier.
        for more information see: qudi.util.datastorage.TextDataStorage.create_header()
        """
        # Create default header as specified in parent TextDataStorage object without column headers
        header = super().create_header(timestamp=timestamp,
                                       metadata=metadata,
                                       notes=notes,
                                       column_headers=column_headers,
                                       column_dtypes=column_dtypes)
        # Append column headers if needed
        if column_headers:
            return f'{header}{format_column_headers(column_headers, self.delimiter)}\n'
        return header

    @staticmethod
    def load_data(file_path):
        """See: DataStorageBase.load_data()

        file_path : str, optional
            Path to file to load data from.
        """
        # Read back metadata
        header, header_lines = get_header_from_file(file_path)
        general, metadata = get_info_from_header(header)
        # Determine dtype specifier from general header section
        dtype = general['column_dtypes']
        if dtype is not None and not isinstance(dtype, type):
            # If dtypes differ, construct a structured array
            if all(dtype[0] == typ for typ in dtype):
                dtype = dtype[0]
            elif str in dtype:
                # handle str type separately since this is (arguably) a bug in numpy.genfromtxt
                dtype = None
            else:
                dtype = [(f'f{col:d}', typ) for col, typ in enumerate(dtype)]
        # Load data from file and skip header
        start_line = header_lines + 1
        if general['column_headers']:
            start_line += 1
        data = np.genfromtxt(file_path,
                             dtype=dtype,
                             comments=general['comments'],
                             delimiter=general['delimiter'],
                             skip_header=start_line)
        return data, metadata, general


class NpyDataStorage(DataStorageBase):
    """Helper class to store (measurement) data on disk as binary .npy file.
    """

    def __init__(self, *, root_dir, **kwargs):
        super().__init__(root_dir=root_dir, **kwargs)

    @property
    def file_extension(self):
        return '.npy'

    def create_header(self, timestamp, dtype, metadata=None, notes=None, column_headers=None):
        """
        """
        # Gather all metadata (both global and locally provided) into a single dict
        metadata = self.get_unified_metadata(metadata)
        return format_header(timestamp,
                             dtype,
                             metadata=metadata,
                             notes=notes,
                             column_headers=column_headers)

    def save_data(self, data, *, metadata=None, notes=None, nametag=None, timestamp=None,
                  column_headers=None, filename=None):
        """Saves a binary file containing the data array.
        Also saves alongside a text file containing the notes, (global) metadata and column headers
        for this data set. The filename of the text file will be the same as for the binary file
        appended by "_metadata".

        For more information, see :meth:`~qudi.util.datastorage.DataStorageBase.save_data`.

        Parameters
        ----------
        column_headers : str or list, optional
            Data column header strings or a single string.
        """
        if timestamp is None:
            timestamp = datetime.now()
        # Construct file name if none is given explicitly
        if filename is None:
            filename = get_timestamp_filename(timestamp=timestamp,
                                              nametag=nametag) + self.file_extension
        # Create filename for separate metadata textfile
        meta_filename = filename.rsplit('.', 1)[0] + '_metadata.txt'

        # Create header
        header = self.create_header(timestamp,
                                    data.dtype,
                                    metadata=metadata,
                                    notes=notes,
                                    column_headers=column_headers)
        # Determine full file path and create containing directories if needed
        file_path = os.path.join(self.root_dir, filename)
        create_dir_for_file(file_path)
        meta_file_path = os.path.join(self.root_dir, meta_filename)
        # Write data and metadata to file. Overwrite silently.
        with open(file_path, 'wb') as file:
            # Write numpy data array in binary format
            np.save(file, data, allow_pickle=False, fix_imports=False)
        with open(meta_file_path, 'w') as file:
            file.write(header)
        return file_path, timestamp, data.shape

    @staticmethod
    def load_data(file_path):
        """
        See :meth:`~DataStorageBase.load_data` for more information.

        Parameters
        ----------
        file_path : str
            Path to the file to load data from.

        Returns
        -------
        np.ndarray
            Data as a numpy array.
        """
        # Load numpy array
        data = np.load(file_path, allow_pickle=False, fix_imports=False)
        # Try to find and load metadata from text file
        metadata_path = file_path.split('.npy')[0] + '_metadata.txt'
        try:
            header, _ = get_header_from_file(metadata_path)
        except FileNotFoundError:
            return data, dict(), dict()
        metadata, general = get_info_from_header(header)
        return data, metadata, general



# -*- coding: utf-8 -*-

"""
This file contains a QWidget very similar to QLineEdit that can also open a file dialog to select
directories and files and display them in the QLineEdit.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['PathLineEdit']

import os
from PySide2 import QtCore, QtWidgets, QtGui
from typing import Optional, Any, List

from qudi.util.paths import get_artwork_dir as _get_artwork_dir


class PathLineEdit(QtWidgets.QWidget):
    """ QLineEdit for editing file system paths directly or via QFileDialog.
    No validation is performed on the entered string.
    Multiple paths are separated by single semicolon, e.g. '<path1>;<path2>;<path3>'

    Specify multiple filters by separating them with double(!) semicolons, e.g.:
    Images (*.png *.xpm *.jpg);;Text files (*.txt);;XML files (*.xml)
    """

    def __init__(self,
                 text: Optional[str] = None,
                 parent: Optional[QtWidgets.QWidget] = None,
                 dialog_caption: Optional[str] = None,
                 root_directory: Optional[str] = None,
                 filters: Optional[str] = None,
                 select_directory: Optional[bool] = False,
                 follow_symlinks: Optional[bool] = False,
                 ) -> None:
        super().__init__(parent=parent)
        self._line_edit = QtWidgets.QLineEdit(text)
        self._tool_button = QtWidgets.QToolButton()
        self._tool_button.setIcon(
            QtGui.QIcon(os.path.join(os.path.join(_get_artwork_dir(), 'icons', 'document-open')))
        )
        self._tool_button.setToolTip('Open file dialog')
        self._tool_button.clicked.connect(self._exec_file_dialog)

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._line_edit)
        layout.addWidget(self._tool_button)
        layout.setStretch(0, 1)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._filters = '' if filters is None else filters
        self._select_directory = bool(select_directory)
        self._follow_symlinks = bool(follow_symlinks)
        self._root_directory = os.path.abspath(os.sep) if root_directory is None else root_directory
        if dialog_caption is None:
            self._dialog_caption = 'Select Directory' if self._select_directory else 'Select Files'
        else:
            self._dialog_caption = dialog_caption

    def __getattr__(self, item: str) -> Any:
        try:
            return getattr(self._line_edit, item)
        except AttributeError:
            pass
        raise AttributeError(f"'{self.__class__.__name__}' object has not attribute '{item}'")

    @property
    def paths(self) -> List[str]:
        paths = (p.strip() for p in self._line_edit.text().split(';'))
        return [p for p in paths if p]

    @QtCore.Slot()
    def _exec_file_dialog(self) -> None:
        self._line_edit.clearFocus()
        dialog = QtWidgets.QFileDialog(parent=self,
                                       caption=self._dialog_caption,
                                       directory=self._root_directory,
                                       filter=self._filters)
        options = QtWidgets.QFileDialog.Option.ReadOnly
        if not self._follow_symlinks:
            options |= QtWidgets.QFileDialog.Option.DontResolveSymlinks
        if self._select_directory:
            dialog.setFileMode(QtWidgets.QFileDialog.FileMode.Directory)
            options |= QtWidgets.QFileDialog.Option.ShowDirsOnly
        else:
            dialog.setFileMode(QtWidgets.QFileDialog.FileMode.ExistingFiles)
        dialog.setOptions(options)
        if dialog.exec_() == QtWidgets.QFileDialog.Accepted:
            paths = dialog.selectedFiles()
            if paths:
                text = ';'.join(p for p in paths if p)
                if text and text != self._line_edit.text():
                    self._line_edit.setText(text)
                    self._line_edit.textEdited.emit(text)
                    self._line_edit.editingFinished.emit()
        dialog.setParent(None)
        dialog.deleteLater()



# -*- coding: utf-8 -*-

"""
QWidgets for configuring the individual module config sections

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['LocalModuleConfigWidget', 'RemoteModuleConfigWidget', 'ModuleConnectorsWidget',
           'ModuleOptionsWidget']

import copy
from PySide2 import QtCore, QtWidgets
from typing import Optional, Iterable, Mapping, Dict, Sequence, Union, Any, Tuple, List

from qudi.core import Connector, ConfigOption
from qudi.core.config.validator import validate_remote_module_config, validate_local_module_config
from qudi.core.config.validator import ValidationError
from qudi.util.widgets.separator_lines import HorizontalLine
from qudi.util.widgets.path_line_edit import PathLineEdit
from qudi.tools.config_editor.custom_widgets import CustomOptionsWidget, CustomConnectorsWidget


class ModuleConnectorsWidget(QtWidgets.QWidget):
    """
    """
    def __init__(self,
                 mandatory_targets: Optional[Mapping[str, Sequence[str]]] = None,
                 optional_targets: Optional[Mapping[str, Sequence[str]]] = None,
                 module_names: Optional[Iterable[str]] = None,
                 config: Optional[Mapping[str, str]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        if mandatory_targets is None:
            self._mandatory_targets = dict()
        else:
            self._mandatory_targets = copy.deepcopy(mandatory_targets)
        if optional_targets is None:
            self._optional_targets = dict()
        else:
            self._optional_targets = copy.deepcopy(optional_targets)
        if set(self._mandatory_targets).intersection(self._optional_targets):
            raise ValueError('Connector names can not be both mandatory AND optional')

        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)

        # Create Caption
        label = QtWidgets.QLabel('Connectors')
        label.setAlignment(QtCore.Qt.AlignCenter)
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label, 0, 0, 1, 2)

        # Keep track of connector editor widgets
        self._connector_editors = dict()
        # Create mandatory connectors
        for row, (name, targets) in enumerate(self._mandatory_targets.items(), 1):
            label, editor = self._make_conn_widgets(name, targets, False)
            layout.addWidget(label, row, 0)
            layout.addWidget(editor, row, 1)
            self._connector_editors[name] = editor
        # Create optional connectors
        offset = len(self._connector_editors) + 1
        for row, (name, targets) in enumerate(self._optional_targets.items(), offset):
            label, editor = self._make_conn_widgets(name, targets, True)
            layout.addWidget(label, row, 0)
            layout.addWidget(editor, row, 1)
            self._connector_editors[name] = editor

        # Add separator
        layout.addWidget(HorizontalLine(), len(self._connector_editors) + 1, 0, 1, 2)

        # Create custom connector editor
        self.custom_connectors_widget = CustomConnectorsWidget(
            forbidden_names=list(self._connector_editors),
            module_names=module_names
        )
        layout.addWidget(self.custom_connectors_widget, len(self._connector_editors) + 2, 0, 1, 2)

        layout.setRowStretch(len(self._connector_editors) + 3, 1)

        self.set_config(config)

    @property
    def config(self) -> Dict[str, Union[None, str]]:
        conn = {name: editor.currentText() for name, editor in self._connector_editors.items()}
        conn.update(self.custom_connectors_widget.config)
        return {name: target if target else None for name, target in conn.items()}

    def set_config(self, config: Union[None, Mapping[str, Union[None, str]]]) -> None:
        if config is None:
            for editor in self._connector_editors.values():
                editor.setCurrentIndex(0)
            self.custom_connectors_widget.set_config(None)
        else:
            cfg = config.copy()
            for name, editor in self._connector_editors.items():
                target = cfg.pop(name, editor.currentText())
                index = max(0, editor.findText(target)) if target else 0
                editor.setCurrentIndex(index)
            # Remaining connectors are custom
            self.custom_connectors_widget.set_config(cfg)

    @staticmethod
    def _make_conn_widgets(name: str,
                           targets: Sequence[str],
                           optional: bool
                           ) -> Tuple[QtWidgets.QLabel, QtWidgets.QComboBox]:
        label = QtWidgets.QLabel(f'{name}:' if optional else f'* {name}:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        label.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        editor = QtWidgets.QComboBox()
        editor.addItem('')
        editor.addItems(targets)
        editor.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        return label, editor


class ModuleOptionsWidget(QtWidgets.QWidget):
    """
    """
    def __init__(self,
                 mandatory_names: Optional[Iterable[str]] = None,
                 optional_names: Optional[Iterable[str]] = None,
                 config: Optional[Mapping[str, Any]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        self._mandatory_names = list() if mandatory_names is None else list(mandatory_names)
        self._optional_names = list() if optional_names is None else list(optional_names)
        if set(self._mandatory_names).intersection(self._optional_names):
            raise ValueError('ConfigOption names can not be both mandatory AND optional')

        layout = QtWidgets.QGridLayout()
        layout.setColumnStretch(1, 1)
        self.setLayout(layout)

        # Create Caption
        label = QtWidgets.QLabel('ConfigOptions')
        label.setAlignment(QtCore.Qt.AlignCenter)
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label, 0, 0, 1, 2)

        # Keep track of option editor widgets
        self._option_editors = dict()
        # Create mandatory options
        for row, name in enumerate(self._mandatory_names, 1):
            label, editor = self._make_option_widgets(name, False)
            layout.addWidget(label, row, 0)
            layout.addWidget(editor, row, 1)
            self._option_editors[name] = editor
        # Create optional options
        offset = len(self._option_editors) + 1
        for row, name in enumerate(self._optional_names, offset):
            label, editor = self._make_option_widgets(name, True)
            layout.addWidget(label, row, 0)
            layout.addWidget(editor, row, 1)
            self._option_editors[name] = editor

        # Add separator
        layout.addWidget(HorizontalLine(), len(self._option_editors) + 1, 0, 1, 2)

        # Create custom options editor
        self.custom_options_widget = CustomOptionsWidget(forbidden_names=list(self._option_editors))
        layout.addWidget(self.custom_options_widget, len(self._option_editors) + 2, 0, 1, 2)

        layout.setRowStretch(len(self._option_editors) + 3, 1)

        self.set_config(config)

    @property
    def config(self) -> Dict[str, Any]:
        cfg = dict()
        for name, editor in self._option_editors.items():
            text = editor.text().strip()
            if text == '':
                # Interpret empty text as None for mandatory options. Skip missing optional options.
                if name in self._optional_names:
                    continue
                else:
                    cfg[name] = None
            else:
                # Try to parse text with eval(). If that fails, interpret text as plain string.
                try:
                    cfg[name] = eval(text)
                except (NameError, SyntaxError, ValueError):
                    cfg[name] = text
        return cfg

    def set_config(self, config: Union[None, Mapping[str, Any]]) -> None:
        if config is None:
            for editor in self._option_editors.values():
                editor.setText('')
            self.custom_options_widget.set_config(None)
        else:
            cfg = config.copy()
            for name, editor in self._option_editors.items():
                try:
                    editor.setText(repr(cfg.pop(name)))
                except:
                    editor.setText('')
            # Remaining options are custom
            self.custom_options_widget.set_config(cfg)

    @staticmethod
    def _make_option_widgets(name: str,
                             optional: bool
                             ) -> Tuple[QtWidgets.QLabel, QtWidgets.QLineEdit]:
        label = QtWidgets.QLabel(f'{name}:' if optional else f'* {name}:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        label.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        editor = QtWidgets.QLineEdit()
        editor.setPlaceholderText('text parsed by eval()')
        return label, editor


class LocalModuleConfigWidget(QtWidgets.QWidget):
    """
    """

    def __init__(self,
                 module_class: str,
                 config_options: Sequence[ConfigOption],
                 connectors: Sequence[Connector],
                 valid_connector_targets: Mapping[str, Sequence[str]],
                 named_modules: Mapping[str, str],
                 config: Optional[Dict[str, Union[str, bool, Dict[str, str], Dict[str, Any]]]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setStretch(2, 1)
        self.setLayout(layout)

        # Module label
        sub_layout = QtWidgets.QGridLayout()
        sub_layout.setColumnStretch(1, 1)
        layout.addLayout(sub_layout)
        label = QtWidgets.QLabel('module.Class:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self._module_label = QtWidgets.QLabel(module_class)
        font = self._module_label.font()
        font.setBold(True)
        self._module_label.setFont(font)
        sub_layout.addWidget(label, 0, 0)
        sub_layout.addWidget(self._module_label, 0, 1)
        # allow_remote flag editor
        label = QtWidgets.QLabel('Allow remote connection:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.allow_remote_checkbox = QtWidgets.QCheckBox()
        self.allow_remote_checkbox.setToolTip(
            'Allow other qudi instances to connect to this module via remote modules server.'
        )
        self.allow_remote_checkbox.toggled.connect(self._validate_and_mark_config)
        sub_layout.addWidget(label, 1, 0)
        sub_layout.addWidget(self.allow_remote_checkbox, 1, 1)

        # Separator
        layout.addWidget(HorizontalLine())

        # Create splitter to spread options and connectors horizontally
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                    QtWidgets.QSizePolicy.Expanding)
        layout.addWidget(self.splitter)
        # Module Connectors editor
        mandatory_targets, optional_targets = self._get_connector_targets(
            connectors=connectors,
            named_modules=named_modules,
            valid_targets=valid_connector_targets
        )
        self.connectors_editor = ModuleConnectorsWidget(mandatory_targets=mandatory_targets,
                                                        optional_targets=optional_targets,
                                                        module_names=list(named_modules))
        self.splitter.addWidget(self.connectors_editor)
        # Module ConfigOptions editor
        self.options_editor = ModuleOptionsWidget(
            mandatory_names=[opt.name for opt in config_options if not opt.optional],
            optional_names=[opt.name for opt in config_options if opt.optional]
        )
        self.splitter.addWidget(self.options_editor)

        self.set_config(config)
        self._validate_and_mark_config()

    @property
    def module_class(self) -> str:
        return self._module_label.text()

    @property
    def config(self) -> Dict[str, Union[str, bool, Dict[str, str], Dict[str, Any]]]:
        return {'module.Class': self.module_class,
                'allow_remote': self.allow_remote_checkbox.isChecked(),
                'options'     : self.options_editor.config,
                'connect'     : self.connectors_editor.config}

    def set_config(self,
                   config: Union[None, Dict[str, Union[str, bool, Dict[str, str], Dict[str, Any]]]]
                   ) -> None:
        if config:
            self.allow_remote_checkbox.setChecked(config.get('allow_remote', False))
            self.options_editor.set_config(config.get('options', dict()))
            self.connectors_editor.set_config(config.get('connect', dict()))
        else:
            self.allow_remote_checkbox.setChecked(False)
            self.options_editor.set_config(None)
            self.connectors_editor.set_config(None)

    @staticmethod
    def _get_connector_targets(connectors: Sequence[Connector],
                               named_modules: Mapping[str, str],
                               valid_targets: Mapping[str, Sequence[str]]
                               ) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
        mandatory_targets = dict()
        optional_targets = dict()
        for conn in connectors:
            targets = [
                name for name, mod in named_modules.items() if mod in valid_targets[conn.name]
            ]
            if conn.optional:
                optional_targets[conn.name] = targets
            else:
                mandatory_targets[conn.name] = targets
        return mandatory_targets, optional_targets

    def validate_config(self) -> None:
        validate_local_module_config(self.config)

    @QtCore.Slot()
    def _validate_and_mark_config(self) -> None:
        try:
            self.validate_config()
        except ValidationError as err:
            print(f'Invalid local module config. Problematic fields: {list(err.relative_path)}')


class RemoteModuleConfigWidget(QtWidgets.QWidget):
    """
    """

    def __init__(self,
                 config: Optional[Mapping[str, Union[str, None]]] = None,
                 parent: Optional[QtWidgets.QWidget] = None
                 ) -> None:
        super().__init__(parent=parent)

        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setColumnStretch(1, 1)
        layout.setRowStretch(5, 1)
        self.setLayout(layout)

        # remote name editor
        label = QtWidgets.QLabel('* Native module name:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.native_name_lineedit = QtWidgets.QLineEdit()
        self.native_name_lineedit.setToolTip('The native module name as configured on the remote '
                                             'host qudi instance to connect to.')
        self.native_name_lineedit.setPlaceholderText('Module name on remote host')
        self.native_name_lineedit.textChanged.connect(self._validate_and_mark_config)
        layout.addWidget(label, 0, 0)
        layout.addWidget(self.native_name_lineedit, 0, 1)

        # remote host editor
        label = QtWidgets.QLabel('* Remote address:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.remote_host_lineedit = QtWidgets.QLineEdit('localhost')
        self.remote_host_lineedit.setToolTip('The IP address of the remote host. Can also be '
                                             '"localhost" for local qudi instances.')
        self.remote_host_lineedit.setPlaceholderText('IP address or "localhost"')
        self.remote_host_lineedit.textChanged.connect(self._validate_and_mark_config)
        layout.addWidget(label, 1, 0)
        layout.addWidget(self.remote_host_lineedit, 1, 1)

        # remote port editor
        label = QtWidgets.QLabel('* Remote port:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.remote_port_spinbox = QtWidgets.QSpinBox()
        self.remote_port_spinbox.setRange(0, 65535)
        self.remote_port_spinbox.setValue(12345)
        self.remote_port_spinbox.setToolTip('Port to reach the remote host on.')
        self.remote_port_spinbox.valueChanged.connect(self._validate_and_mark_config)
        layout.addWidget(label, 2, 0)
        layout.addWidget(self.remote_port_spinbox, 2, 1)

        # certfile editor
        label = QtWidgets.QLabel('Certificate file:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.certfile_lineedit = PathLineEdit(dialog_caption='Select SSL Certificate File',
                                              follow_symlinks=True)
        self.certfile_lineedit.setPlaceholderText('No certificate')
        self.certfile_lineedit.setToolTip(
            'SSL certificate file path for the remote module connection'
        )
        self.certfile_lineedit.textChanged.connect(self._validate_and_mark_config)
        layout.addWidget(label, 3, 0)
        layout.addWidget(self.certfile_lineedit, 3, 1)

        # keyfile editor
        label = QtWidgets.QLabel('Key file:')
        label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
        self.keyfile_lineedit = PathLineEdit(dialog_caption='Select SSL Key File',
                                             follow_symlinks=True)
        self.keyfile_lineedit.setPlaceholderText('No key')
        self.keyfile_lineedit.setToolTip('SSL key file path for the remote module server')
        self.keyfile_lineedit.textChanged.connect(self._validate_and_mark_config)
        layout.addWidget(label, 4, 0)
        layout.addWidget(self.keyfile_lineedit, 4, 1)

        self.set_config(config)
        self._validate_and_mark_config()

    @property
    def config(self) -> Dict[str, Union[None, int, str]]:
        native_module_name = self.native_name_lineedit.text()
        host = self.remote_host_lineedit.text()
        cfg = {'native_module_name': native_module_name if native_module_name else None,
               'address'           : host if host else None,
               'port'              : self.remote_port_spinbox.value()}
        try:
            cfg['certfile'] = self.certfile_lineedit.paths[0]
        except IndexError:
            pass
        try:
            cfg['keyfile'] = self.keyfile_lineedit.paths[0]
        except IndexError:
            pass
        return cfg

    def set_config(self, config: Union[None, Dict[str, Union[None, int, str]]]) -> None:
        if config:
            native_module_name = config.get('native_module_name', None)
            host = config.get('address', None)
            port = config.get('port', None)
            try:
                certfile = config['certfile']
                keyfile = config['keyfile']
            except KeyError:
                certfile = keyfile = ''
            if certfile is None or keyfile is None:
                certfile = keyfile = ''
            self.remote_host_lineedit.setText(host if host else '')
            self.remote_port_spinbox.setValue(port if isinstance(port, int) else 12345)
            self.native_name_lineedit.setText(native_module_name if native_module_name else '')
            self.certfile_lineedit.setText(certfile)
            self.certfile_lineedit.setText(keyfile)
        else:
            self.remote_host_lineedit.setText('')
            self.remote_port_spinbox.setValue(12345)
            self.native_name_lineedit.setText('')
            self.certfile_lineedit.setText('')
            self.certfile_lineedit.setText('')

    def validate_config(self) -> None:
        validate_remote_module_config(self.config)

    @QtCore.Slot()
    def _validate_and_mark_config(self) -> None:
        try:
            self.validate_config()
        except ValidationError as err:
            print(f'Invalid remote module config. Problematic fields: {list(err.relative_path)}')



# -*- coding: utf-8 -*-
""" This module contains the

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import logging
import subprocess
import jupyter_client.kernelspec
from PySide2 import QtCore, QtWidgets
from qtconsole.manager import QtKernelManager
from collections.abc import Mapping, Sequence, Set

from qudi.core.statusvariable import StatusVar
from qudi.core.threadmanager import ThreadManager
from qudi.util.paths import get_main_dir, get_default_config_dir
from qudi.core.gui.main_gui.errordialog import ErrorDialog
from qudi.core.gui.main_gui.mainwindow import QudiMainWindow
from qudi.core.module import GuiBase
from qudi.core.logger import get_signal_handler

try:
    from git import Repo, InvalidGitRepositoryError
except ImportError:
    Repo = None


class QudiMainGui(GuiBase):
    """
    This class provides a GUI to the qudi main application object.
    """
    # status vars
    _console_font_size = StatusVar(name='console_font_size', default=10)
    _show_error_popups = StatusVar(name='show_error_popups', default=True)

    def __init__(self, *args, **kwargs):
        """Create an instance of the module.

          @param object manager:
          @param str name:
          @param dict config:
        """
        super().__init__(*args, **kwargs)
        self.error_dialog = None
        self.mw = None
        self._has_console = False  # Flag indicating if an IPython console is available

    def on_activate(self):
        """Activation method called on change to active state.

        This method creates the Manager main window.
        """
        # Create main window and restore position
        self.mw = QudiMainWindow(debug_mode=self._qudi_main.debug_mode)
        self._restore_window_geometry(self.mw)
        # Create error dialog for error message popups
        self.error_dialog = ErrorDialog()
        self.error_dialog.set_enabled(self._show_error_popups)

        # Get qudi version number and configure statusbar and "about qudi" dialog
        version = self.get_qudi_version()
        if isinstance(version, str):
            self.mw.about_qudi_dialog.version_label.setText('version {0}'.format(version))
            self.mw.version_label.setText(
                '<a style=\"color: cyan;\"> version {0} </a>  configured from {1}'
                ''.format(version, self._qudi_main.configuration.file_path))
        else:
            self.mw.about_qudi_dialog.version_label.setText(
                '<a href=\"https://github.com/Ulm-IQO/qudi/commit/{0}\" style=\"color: cyan;\"> {0}'
                ' </a>, on branch {1}.'.format(version[0], version[1]))
            self.mw.version_label.setText(
                '<a href=\"https://github.com/Ulm-IQO/qudi/commit/{0}\" style=\"color: cyan;\"> {0}'
                ' </a>, on branch {1}, configured from {2}'
                ''.format(version[0], version[1], self._qudi_main.configuration.file_path))

        self._connect_signals()

        self.keep_settings()
        self.update_configured_modules()
        self.update_config_widget()

        # IPython console widget
        self.start_jupyter_widget()

        # Configure thread widget
        self.mw.threads_widget.setModel(ThreadManager.instance())

        # Configure remotemodules widget
        self._init_remote_modules_widget()

        self.reset_default_layout()
        self.show()

    def on_deactivate(self):
        """Close window and remove connections.
        """
        self._disconnect_signals()
        self.stop_jupyter_widget()
        self._save_window_geometry(self.mw)
        self.mw.close()

    def _connect_signals(self):
        get_signal_handler().sigRecordLogged.connect(self.handle_log_record, QtCore.Qt.QueuedConnection)
        qudi_main = self._qudi_main
        # Connect up the main windows actions
        self.mw.action_quit.triggered.connect(qudi_main.prompt_quit, QtCore.Qt.QueuedConnection)
        self.mw.action_load_configuration.triggered.connect(self.load_configuration)
        self.mw.action_reload_qudi.triggered.connect(
            qudi_main.prompt_restart, QtCore.Qt.QueuedConnection)
        self.mw.action_open_configuration_editor.triggered.connect(self.new_configuration)
        self.mw.action_load_all_modules.triggered.connect(
            qudi_main.module_manager.start_all_modules)
        self.mw.action_view_default.triggered.connect(self.reset_default_layout)
        # Connect signals from manager
        qudi_main.configuration.sigConfigChanged.connect(self.update_config_widget)
        qudi_main.module_manager.sigManagedModulesChanged.connect(self.update_configured_modules)
        qudi_main.module_manager.sigModuleStateChanged.connect(self.update_module_state)
        qudi_main.module_manager.sigModuleAppDataChanged.connect(self.update_module_app_data)
        # Settings dialog
        self.mw.settings_dialog.accepted.connect(self.apply_settings)
        self.mw.settings_dialog.rejected.connect(self.keep_settings)
        self.error_dialog.disable_checkbox.clicked.connect(self._error_dialog_enabled_changed)
        # Modules list
        self.mw.module_widget.sigActivateModule.connect(qudi_main.module_manager.activate_module)
        self.mw.module_widget.sigReloadModule.connect(qudi_main.module_manager.reload_module)
        self.mw.module_widget.sigDeactivateModule.connect(
            qudi_main.module_manager.deactivate_module)
        self.mw.module_widget.sigCleanupModule.connect(
            qudi_main.module_manager.clear_module_app_data)

    def _disconnect_signals(self):
        qudi_main = self._qudi_main
        # Disconnect the main windows actions
        self.mw.action_quit.triggered.disconnect()
        self.mw.action_load_configuration.triggered.disconnect()
        self.mw.action_reload_qudi.triggered.disconnect()
        self.mw.action_open_configuration_editor.triggered.disconnect()
        self.mw.action_load_all_modules.triggered.disconnect()
        self.mw.action_view_default.triggered.disconnect()
        # Disconnect signals from manager
        qudi_main.configuration.sigConfigChanged.disconnect(self.update_config_widget)
        qudi_main.module_manager.sigManagedModulesChanged.disconnect(self.update_configured_modules)
        qudi_main.module_manager.sigModuleStateChanged.disconnect(self.update_module_state)
        qudi_main.module_manager.sigModuleAppDataChanged.disconnect(self.update_module_app_data)
        # Settings dialog
        self.mw.settings_dialog.accepted.disconnect()
        self.mw.settings_dialog.rejected.disconnect()
        self.error_dialog.disable_checkbox.clicked.disconnect()
        # Modules list
        self.mw.module_widget.sigActivateModule.disconnect()
        self.mw.module_widget.sigReloadModule.disconnect()
        self.mw.module_widget.sigDeactivateModule.disconnect()
        self.mw.module_widget.sigCleanupModule.disconnect()

        get_signal_handler().sigRecordLogged.disconnect(self.handle_log_record)

    def _init_remote_modules_widget(self):
        remote_server = self._qudi_main.remote_modules_server
        # hide remote modules menu action if RemoteModuleServer is not available
        if remote_server is None:
            self.mw.remote_widget.setVisible(False)
            self.mw.remote_dockwidget.setVisible(False)
            self.mw.action_view_remote.setVisible(False)
        else:
            server_config = self._qudi_main.configuration['remote_modules_server']
            host = server_config['address']
            port = server_config['port']
            self.mw.remote_widget.setVisible(True)
            self.mw.remote_widget.server_label.setText(f'Server URL: rpyc://{host}:{port}/')
            self.mw.remote_widget.shared_module_listview.setModel(
                remote_server.service.shared_modules
            )

    def show(self):
        """Show the window and bring it to the top.
        """
        self.mw.show()
        self.mw.activateWindow()
        self.mw.raise_()

    def reset_default_layout(self):
        """
        Return the dockwidget layout and visibility to its default state.
        """
        self.mw.config_dockwidget.setVisible(False)
        self.mw.console_dockwidget.setVisible(self._has_console)
        self.mw.remote_dockwidget.setVisible(False)
        self.mw.threads_dockwidget.setVisible(False)
        self.mw.log_dockwidget.setVisible(True)

        self.mw.config_dockwidget.setFloating(False)
        self.mw.console_dockwidget.setFloating(False)
        self.mw.remote_dockwidget.setFloating(False)
        self.mw.threads_dockwidget.setFloating(False)
        self.mw.log_dockwidget.setFloating(False)

        self.mw.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.mw.config_dockwidget)
        self.mw.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.mw.log_dockwidget)
        self.mw.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.mw.remote_dockwidget)
        self.mw.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.mw.threads_dockwidget)
        self.mw.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.mw.console_dockwidget)

        self.mw.action_view_console.setChecked(self._has_console)
        self.mw.action_view_console.setVisible(self._has_console)
        return

    def handle_log_record(self, entry):
        """
        Show an error popup if the log entry is error level and above.

        @param logging.LogRecord entry: log record as returned from logging module
        """
        if entry.levelname in ('error', 'critical'):
            self.error_dialog.new_error(entry)
        return

    def start_jupyter_widget(self):
        """Starts a qudi IPython kernel in a separate process and connects it to the console widget.
        """
        self._has_console = False
        try:
            # Create and start kernel process
            kernel_manager = QtKernelManager(kernel_name='qudi', autorestart=False)
            # kernel_manager.kernel.gui = 'qt4'
            kernel_manager.start_kernel()

            # create kernel client and connect to console widget
            banner = 'This is an interactive IPython console. A reference to the running qudi ' \
                     'instance can be accessed via "qudi". View the current namespace with dir().\n' \
                     'Go, play.\n'
            self.mw.console_widget.banner = banner
            self.mw.console_widget.font_size = self._console_font_size
            self.mw.console_widget.reset_font()
            self.mw.console_widget.set_default_style(colors='linux')
            kernel_client = kernel_manager.client()
            kernel_client.hb_channel.time_to_dead = 10.0
            kernel_client.hb_channel.kernel_died.connect(self.kernel_died_callback)
            kernel_client.start_channels()
            self.mw.console_widget.kernel_manager = kernel_manager
            self.mw.console_widget.kernel_client = kernel_client
            self._has_console = True
            self.log.info('IPython kernel for qudi main GUI successfully started.')
        except jupyter_client.kernelspec.NoSuchKernel:
            self.log.warn(
                'Qudi IPython kernelspec not installed.\n'
                'IPython console and jupyter notebook integration not available.\n'
                'Run "qudi-install-kernel" from within the qudi Python environment to fix this. '
            )
        except:
            self.log.exception(
                'Exception while trying to start IPython kernel for qudi main GUI. Qudi IPython '
                'console not available.'
            )

    @QtCore.Slot()
    def kernel_died_callback(self):
        """
        """
        try:
            self.mw.console_widget.kernel_client.stop_channels()
        except:
            pass
        if self._has_console:
            self._has_console = False
            self.log.error(
                'Qudi IPython kernel has unexpectedly died. This can be caused by a corrupt qudi '
                'kernelspec installation. Try to run "qudi-install-kernel" from within the qudi '
                'Python environment and restart qudi.'
            )

    def stop_jupyter_widget(self):
        """Stops the qudi IPython kernel process and detaches it from the console widget.
        """
        try:
            self.mw.console_widget.kernel_client.stop_channels()
        except:
            self.log.exception('Exception while trying to shutdown qudi IPython client:')
        try:
            self.mw.console_widget.kernel_manager.shutdown_kernel()
        except:
            self.log.exception('Exception while trying to shutdown qudi IPython kernel:')
        self._has_console = False
        self.log.info('IPython kernel process for qudi main GUI has shut down.')

    def keep_settings(self):
        """Write old values into settings dialog.
        """
        self.mw.settings_dialog.font_size_spinbox.setValue(self._console_font_size)
        self.mw.settings_dialog.show_error_popups_checkbox.setChecked(self._show_error_popups)

    def apply_settings(self):
        """Apply values from settings dialog.
        """
        # Console font size
        font_size = self.mw.settings_dialog.font_size_spinbox.value()
        self.mw.console_widget.font_size = font_size
        self.mw.console_widget.reset_font()
        self._console_font_size = font_size

        # Error popups
        error_popups = self.mw.settings_dialog.show_error_popups_checkbox.isChecked()
        self.error_dialog.set_enabled(error_popups)
        self._show_error_popups = error_popups

    @QtCore.Slot()
    def _error_dialog_enabled_changed(self):
        """Callback for the error dialog disable checkbox.
        """
        self._show_error_popups = self.error_dialog.enabled
        self.mw.settings_dialog.show_error_popups_checkbox.setChecked(self._show_error_popups)

    @QtCore.Slot(object)
    def update_config_widget(self, config=None):
        """Clear and refill the tree widget showing the configuration.
        """
        if config is None:
            config = self._qudi_main.configuration
        self.mw.config_widget.set_config(config.config_map)

    @QtCore.Slot(dict)
    def update_configured_modules(self, modules=None):
        """Clear and refill the module list widget.
        """
        if modules is None:
            modules = self._qudi_main.module_manager
        self.mw.module_widget.update_modules(modules)

    @QtCore.Slot(str, str, str)
    def update_module_state(self, base, name, state):
        self.mw.module_widget.update_module_state(base, name, state)
        return

    @QtCore.Slot(str, str, bool)
    def update_module_app_data(self, base, name, exists):
        self.mw.module_widget.update_module_app_data(base, name, exists)

    def get_qudi_version(self):
        """Try to determine the software version in case the program is in a git repository.
        """
        # Try to get repository information if qudi has been checked out as git repo
        if Repo is not None:
            try:
                repo = Repo(os.path.dirname(get_main_dir()))
                branch = repo.active_branch
                rev = str(repo.head.commit)
                return rev, str(branch)
            except InvalidGitRepositoryError:
                pass
            except:
                self.log.exception('Unexpected error while trying to get git repo:')

        # Try to get qudi.core version number
        try:
            from qudi.core import __version__
            return __version__
        except:
            self.log.exception('Unexpected error while trying to get qudi version:')
        return 'unknown'

    def load_configuration(self):
        """Ask the user for a file where the configuration should be loaded from.
        """
        filename = QtWidgets.QFileDialog.getOpenFileName(self.mw,
                                                         'Load Configuration',
                                                         get_default_config_dir(True),
                                                         'Configuration files (*.cfg)')[0]
        if filename:
            reply = QtWidgets.QMessageBox.question(
                self.mw,
                'Restart',
                'Do you want to restart to use the configuration?\n'
                'Choosing "No" will use the selected config file for the next start of Qudi.',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel,
                QtWidgets.QMessageBox.No
            )
            if reply == QtWidgets.QMessageBox.Cancel:
                return
            self._qudi_main.configuration.set_default_path(filename)
            if reply == QtWidgets.QMessageBox.Yes:
                self._qudi_main.restart()

    def new_configuration(self):
        """Prompt the user to open the graphical config editor in a subprocess in order to
        edit/create config files for qudi.
        """
        reply = QtWidgets.QMessageBox.question(
                self.mw,
                'Open Configuration Editor',
                'Do you want open the graphical qudi configuration editor to create or edit qudi '
                'config files?\n',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.Yes
        )
        if reply == QtWidgets.QMessageBox.Yes:
            process = subprocess.Popen(args=[sys.executable, '-m', 'tools.config_editor'],
                                       close_fds=False,
                                       env=os.environ.copy(),
                                       stdin=sys.stdin,
                                       stdout=sys.stdout,
                                       stderr=sys.stderr)



# -*- coding: utf-8 -*-
"""
This module provides functionality for linear transformations of cartesian coordinate systems.

Copyright (c) 2022, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-iqo-modules/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['LinearTransformation', 'LinearTransformation3D', 'LinearTransformation2D']

import numpy as np
from typing import Sequence, Optional, Union, Tuple

from qudi.util.helpers import is_integer
from qudi.util.math import normalize

class LinearTransformation:
    """Linear transformation for N-dimensional cartesian coordinates."""

    def __init__(self,
                 matrix: Optional[Sequence[Sequence[float]]] = None,
                 dimensions: Optional[int] = None
                 ) -> None:
        super().__init__()

        if matrix is not None:
            self._matrix = np.array(matrix, dtype=float)
            if self._matrix.ndim != 2:
                raise ValueError('LinearTransformation matrix must be 2-dimensional')
            if self._matrix.shape[0] != self._matrix.shape[1]:
                raise ValueError('LinearTransformation matrix must be square')
        elif dimensions is not None:
            if not is_integer(dimensions):
                raise TypeError(f'LinearTransformation dimensions must be integer type. '
                                f'Received {type(dimensions)} instead.')
            if dimensions < 1:
                raise ValueError(f'LinearTransformation dimensions must >= 1. '
                                 f'Received {dimensions:d} instead.')
            self._matrix = np.eye(dimensions + 1, dimensions + 1)
        else:
            raise ValueError('Must either provide homogenous transformation matrix or number of '
                             'dimensions')

    def __call__(self,
                 nodes: Union[Sequence[float], Sequence[Sequence[float]]],
                 invert: Optional[bool] = False
                 ) -> np.ndarray:
        """Transforms any single node (vector) or sequence of nodes according to the
        preconfigured matrix.
        Tries to perform the inverse transform if the optional argument invert is True.
        """
        nodes = np.squeeze(np.asarray(nodes))
        node_dim = np.ndim(nodes)
        matrix = self.inverse if invert else self._matrix
        if node_dim == 2:
            nodes = np.vstack([nodes.T, np.full(nodes.shape[0], 1)])
            return np.matmul(matrix, nodes)[:self.dimensions, :].T
        elif node_dim == 1:
            nodes = np.append(nodes, 1)
            return np.matmul(matrix, nodes)[:self.dimensions]
        raise ValueError('nodes to transform must either be 1D or 2D array')

    @property
    def matrix(self) -> np.ndarray:
        """Returns a copy of the currently configured homogenous transformation matrix
        (including translation).
        """
        return self._matrix.copy()

    @property
    def inverse(self) -> np.ndarray:
        """Returns a copy of the inverse of the currently configured homogenous transformation
        matrix.
        """
        return np.linalg.inv(self._matrix)

    @property
    def dimensions(self) -> int:
        """Returns the number of dimensions of the coordinate system to transform.
        The homogenous transformation matrix is a square matrix with (dimensions + 1) rows/columns.
        """
        return self._matrix.shape[0] - 1

    def add_transform(self, matrix: Sequence[Sequence[float]]) -> None:
        """Multiply a given homogenous transformation matrix (including translation) onto the
        current trasnformation matrix.
        """
        matrix = np.asarray(matrix, dtype=float)
        if matrix.shape != self._matrix.shape:
            raise ValueError(f'LinearTransformation.add_transform expects a homogenious '
                             f'transformation matrix with the same shape as '
                             f'LinearTransformation.matrix {self._matrix.shape}. '
                             f'Received {matrix.shape} instead.')
        self._matrix = np.matmul(matrix, self._matrix)

    def translate(self, *args: float) -> None:
        """Adds a translation to the transformation. Must provide a displacement argument for
        each dimension.
        """
        dim = self.dimensions
        if len(args) != dim:
            raise ValueError(f'LinearTransformation.translate requires as many arguments as '
                             f'number of dimensions ({dim:d})')
        translate_matrix = np.asarray(np.diag([1]*(dim+1)), dtype=float)
        translate_matrix[:-1, -1] = args
        self.add_transform(translate_matrix)

    def scale(self, *args: float) -> None:
        """Adds scaling to the transformation. Must provide a scale factor argument for each
        dimension.
        """
        diagonal = np.ones(self._matrix.shape[0], dtype=float)
        if len(args) == 1:
            diagonal[:-1] *= args[0]
        elif len(args) == self.dimensions:
            diagonal[:-1] *= args
        else:
            raise ValueError(f'LinearTransformation.scale requires either a single argument or as '
                             f'many arguments as number of dimensions ({self.dimensions:d})')
        scale_matrix = np.diag(diagonal)
        self.add_transform(scale_matrix)

    def rotate(self, *args, **kwargs) -> None:
        """Adds a rotation to the transformation. Must provide a rotation angle argument for each
        axis (dimension).
        """
        raise NotImplementedError('Arbitrary rotation transformation not implemented yet')

    def from_support_vectors(self):
        # todo
        pass

class LinearTransformation3D(LinearTransformation):
    """Linear transformation for 3D cartesian coordinates."""

    _Vector = Tuple[float, float, float, float]
    _TransformationMatrix = Tuple[_Vector, _Vector, _Vector, _Vector]

    def __init__(self, matrix: Optional[_TransformationMatrix] = None) -> None:
        super().__init__(matrix=matrix, dimensions=3)

    def rotate(self,
               x_angle: Optional[float] = 0,
               y_angle: Optional[float] = 0,
               z_angle: Optional[float] = 0
               ) -> None:
        """Adds a rotation to the transformation. Can provide a rotation angle (in rad) around
        each of the 3 axes (x, y, z).
        """
        sin_a = np.sin(x_angle)
        cos_a = np.cos(x_angle)
        sin_b = np.sin(y_angle)
        cos_b = np.cos(y_angle)
        sin_c = np.sin(z_angle)
        cos_c = np.cos(z_angle)
        rot_matrix = np.array([
            [cos_b * cos_c, sin_a * sin_b * cos_c - cos_a * sin_c, cos_a * sin_b * cos_c + sin_a * sin_c, 0],
            [cos_b * sin_c, sin_a * sin_b * sin_c + cos_a * cos_c, cos_a * sin_b * sin_c - sin_a * cos_c, 0],
            [-sin_b,        sin_a * cos_b,                         cos_a * cos_b,                         0],
            [0,             0,                                     0,                                     1]
        ])
        self.add_transform(rot_matrix)

    def translate(self,
                  dx: Optional[float] = 0,
                  dy: Optional[float] = 0,
                  dz: Optional[float] = 0
                  ) -> None:
        """Adds a translation to the transformation. Can provide a displacement for each of the 3
        axes (x, y, z).
        """
        return super().translate(dx, dy, dz)

    def scale(self,
              sx: Optional[float] = 1,
              sy: Optional[float] = 1,
              sz: Optional[float] = 1
              ) -> None:
        """Adds scaling to the transformation. Can provide a scale factor for each of the 3 axes
        (x, y, z).
        """
        return super().scale(sx, sy, sz)

    def add_rotation(self, matrix) -> None:
        """
        Add a rotation given by 3x3 matrix. Pad the array to represent this rotation plus a zero translation.
        
        Parameters
        ----------
        matrix
        """
        rot_matrix = np.pad(matrix, [(0, 1), (0, 1)])
        rot_matrix[-1,-1] = 1

        self.add_transform(rot_matrix)


class LinearTransformation2D(LinearTransformation):
    """Linear transformation for 2D cartesian coordinates."""

    _Vector = Tuple[float, float, float]
    _TransformationMatrix = Tuple[_Vector, _Vector, _Vector]

    def __init__(self, matrix: Optional[_TransformationMatrix] = None) -> None:
        super().__init__(matrix=matrix, dimensions=2)

    def rotate(self, angle: float) -> None:
        """ Adds a rotation to the transformation. Given angle (in rad) will rotate around origin
        counter-clockwise.
        """
        cos = np.cos(angle)
        sin = np.sin(angle)
        rot_matrix = np.array([
            [cos, -sin, 0],
            [sin,  cos, 0],
            [0  ,    0, 1]
        ])
        self.add_transform(rot_matrix)

    def translate(self, dx: Optional[float] = 0, dy: Optional[float] = 0) -> None:
        """Adds a translation to the transformation. Can provide a displacement for each of the 2
        axes (x, y).
        """
        return super().translate(dx, dy)

    def scale(self, sx: Optional[float] = 1, sy: Optional[float] = 1) -> None:
        """Adds scaling to the transformation. Can provide a scale factor for each of the 2 axes
        (x, y).
        """
        return super().scale(sx, sy)

# todo: integrate in math.py or linear_transform.py

def find_changing_axes(points: np.ndarray) -> np.ndarray:
    """
    From a set of column vectors, find the axes which do not have the same coordinate for all vectors.

    Parameters
    ----------
    points : numpy.ndarray
        Each row represents a vector.

    Returns
    -------
    numpy.ndarray
        Array of the indices of the changing axes.
    """
    num_axes = len(points[0])
    axes_changing_p = np.zeros(num_axes, dtype=np.bool_)
    # substract each
    for axis in range(num_axes):
        elements = points[:, axis]
        for ii, element in enumerate(points[:, axis]):
            d_elements = np.abs(element - elements[ii+1:])
            if np.any(d_elements > 0):
                axes_changing_p[axis] = True
                break
    return axes_changing_p

def compute_reduced_vectors(points: np.ndarray) -> np.ndarray:
    """
    From a set of column vectors, keep only the columns (axes) that are not constant coordinates
    for all vectors.

    Parameters
    ----------
    points : numpy.ndarray
        Each row represents a vector.

    Returns
    -------
    numpy.ndarray
        Reduced array containing only the non-constant axes.
    """
    axes_changing_p = find_changing_axes(points)
    return points[:, axes_changing_p]

def compute_rotation_matrix_to_plane(v0: np.ndarray, v1: np.ndarray, v2: np.ndarray, ez=[0,0,1]) -> np.ndarray:
    """
    Find the rotation matrix that transforms a plane given by three support vectors onto the z plane.
    This rotation is around the origin of the coordinate system.
    Optionally, define the target plane by its normal vector.

    Parameters
    ----------
    v0 : numpy.ndarray
        Support vector 0.
    v1 : numpy.ndarray
        Support vector 1.
    v2 : numpy.ndarray
        Support vector 2.
    ez : numpy.ndarray, optional
        Normal vector defining the target plane. By default, this is the z plane.

    Returns
    -------
    numpy.ndarray
        3x3 rotation matrix.

    Notes
    -----
    See the math here: https://en.wikipedia.org/wiki/Rodrigues'_rotation_formula:return:
    """
    if len(v0) != 3 or len(v1) != 3 or len(v2) != 3:
        raise ValueError('The support vectors should have a length of 3.')
    s0 = v1 - v0
    s1 = v2 - v0
    ez = np.asarray(ez)

    normal_plane_vec = normalize(np.cross(s0, s1))[0]
    rot_axis = normalize(np.cross(normal_plane_vec, ez))[0]

    kx, ky, kz = rot_axis[0], rot_axis[1], rot_axis[2]
    k_mat = np.array([[0.0, -kz, ky], [kz, 0.0, -kx], [-ky, kx, 0.0]])

    theta = -np.arccos(np.dot(normal_plane_vec, ez))
    if theta > np.pi/2 or theta < -np.pi/2:
        theta = -(np.pi-theta)

    return np.eye(3) + np.sin(theta) * k_mat + (1 - np.cos(theta)) * np.matmul(k_mat, k_mat)



# -*- coding: utf-8 -*-
"""
This file contains the Qudi tools for remote module sharing via rpyc server.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('get_remote_module_instance', 'BaseServer', 'RemoteModulesServer', 'QudiNamespaceServer')

import ssl
import rpyc
import weakref
from PySide2 import QtCore
from urllib.parse import urlparse
from rpyc.utils.authenticators import SSLAuthenticator

from qudi.util.mutex import Mutex
from qudi.core.logger import get_logger
from qudi.core.services import RemoteModulesService, QudiNamespaceService

logger = get_logger(__name__)


def get_remote_module_instance(remote_url, certfile=None, keyfile=None, protocol_config=None):
    """Helper method to retrieve a remote module instance via rpyc from a qudi RemoteModuleServer.

    Parameters
    ----------
    remote_url : str
        The URL of the remote qudi module.
    certfile : str
        Certificate file path for the request.
    keyfile : str
        Key file path for the request.
    protocol_config : dict, optional
        Configuration options for rpyc.ssl_connect.

    Returns
    -------
    object or None
        The requested qudi module instance. Returns None if the request failed.
    """
    parsed = urlparse(remote_url)
    if protocol_config is None:
        protocol_config = {'allow_all_attrs': True,
                           'allow_setattr': True,
                           'allow_delattr': True,
                           'allow_pickle': True,
                           'sync_request_timeout': 3600}
    if certfile is not None and keyfile is not None:
        connection = rpyc.ssl_connect(host=parsed.hostname,
                                      port=parsed.port,
                                      config=protocol_config,
                                      certfile=certfile,
                                      keyfile=keyfile)
    else:
        connection = rpyc.connect(host=parsed.hostname,
                                  port=parsed.port,
                                  config=protocol_config,)
    logger.debug(f'get_remote_module_instance has protocol_config {protocol_config}')
    return connection.root.get_module_instance(parsed.path.replace('/', ''))


class _ServerRunnable(QtCore.QObject):
    """QObject containing the actual long-running code to execute in a separate thread for qudi
    RPyC servers.
    """

    def __init__(self, service, host, port, certfile=None, keyfile=None, protocol_config=None,
                 ssl_version=None, cert_reqs=None, ciphers=None):
        super().__init__()

        self.service = service
        self.server = None

        self.host = host
        self.port = port
        self.certfile = certfile
        self.keyfile = keyfile
        if protocol_config is None:
            self.protocol_config = {'allow_all_attrs': True,
                                    'allow_setattr': True,
                                    'allow_delattr': True,
                                    'allow_pickle': True,
                                    'sync_request_timeout': 3600}
        else:
            self.protocol_config = protocol_config
        self.ssl_version = ssl.PROTOCOL_TLSv1_2 if ssl_version is None else ssl_version
        self.cert_reqs = ssl.CERT_REQUIRED if cert_reqs is None else cert_reqs
        self.ciphers = 'EECDH+AESGCM:EDH+AESGCM:AES256+EECDH:AES256+EDH' if ciphers is None else ciphers

    @QtCore.Slot()
    def run(self):
        """Start the RPyC server.
        """
        if self.certfile is not None and self.keyfile is not None:
            authenticator = SSLAuthenticator(certfile=self.certfile,
                                             keyfile=self.keyfile,
                                             cert_reqs=self.cert_reqs,
                                             ssl_version=self.ssl_version,
                                             ciphers=self.ciphers)
        else:
            authenticator = None

        try:
            self.server = rpyc.ThreadedServer(self.service,
                                              hostname=self.host,
                                              port=self.port,
                                              protocol_config=self.protocol_config,
                                              authenticator=authenticator)
            logger.info(f'Starting RPyC server "{self.thread().objectName()}" on '
                        f'[{self.host}]:{self.port:d}')
            logger.debug(f'{self.thread().objectName()}: '
                         f'protocol_config is {self.protocol_config}, '
                         f'authenticator is {authenticator}')
            self.server.start()
        except:
            logger.exception(f'Error during start of RPyC Server "{self.thread().objectName()}":')
            self.server = None

    @QtCore.Slot()
    def stop(self):
        """Stop the RPyC server.
        """
        if self.server is not None:
            try:
                self.server.close()
                logger.info(f'Stopped RPyC server on [{self.host}]:{self.port:d}')
            except:
                logger.exception(
                    f'Exception while trying to stop RPyC server on [{self.host}]:{self.port:d}'
                )
            finally:
                self.server = None


class BaseServer(QtCore.QObject):
    """Contains a threaded RPyC server providing given service.
    USE SSL AUTHENTICATION WHEN LISTENING ON ANYTHING ELSE THAN "localhost"/127.0.0.1.
    Actual RPyC server runs in a QThread.
    """

    def __init__(self, qudi, service_instance, name, host, port, certfile=None,
                 keyfile=None, protocol_config=None, ssl_version=None, cert_reqs=None,
                 ciphers=None, parent=None):
        """
        Parameters
        ----------
        port : int
            Port number the RPyC server should listen to.
        """
        super().__init__(parent=parent)

        self.__qudi_ref = weakref.ref(qudi)
        self._thread_lock = Mutex()

        self.service = service_instance
        self._name = name
        self._server = _ServerRunnable(service=service_instance,
                                       host=host,
                                       port=port,
                                       certfile=certfile,
                                       keyfile=keyfile,
                                       protocol_config=protocol_config,
                                       ssl_version=ssl_version,
                                       cert_reqs=cert_reqs,
                                       ciphers=ciphers)

    @property
    def server(self):
        return self._server.server

    @property
    def is_running(self):
        with self._thread_lock:
            return self._server.server is not None

    @property
    def _qudi(self):
        qudi = self.__qudi_ref()
        if qudi is None:
            raise RuntimeError('Dead qudi application reference encountered')
        return qudi

    @property
    def _thread_manager(self):
        manager = self._qudi.thread_manager
        if manager is None:
            raise RuntimeError('No thread manager initialized in qudi application')
        return manager

    @property
    def _module_manager(self):
        manager = self._qudi.module_manager
        if manager is None:
            raise RuntimeError('No module manager initialized in qudi application')
        return manager

    @QtCore.Slot()
    def start(self):
        """Start the RPyC server.
        """
        with self._thread_lock:
            if self.server is None:
                thread = self._thread_manager.get_new_thread(self._name)
                self._server.moveToThread(thread)
                thread.started.connect(self._server.run)
                thread.start()
            else:
                logger.warning(f'RPyC server "{self._name}" is already running.')

    @QtCore.Slot()
    def stop(self):
        """Stop the RPyC server.
        """
        with self._thread_lock:
            if self.server is not None:
                try:
                    self._server.stop()
                finally:
                    thread_manager = self._thread_manager
                    thread_manager.quit_thread(self._name)
                    thread_manager.join_thread(self._name, time=5)


class RemoteModulesServer(BaseServer):
    """
    """

    def __init__(self, force_remote_calls_by_value=False, **kwargs):
        kwargs['service_instance'] = RemoteModulesService(
            force_remote_calls_by_value=force_remote_calls_by_value
        )
        super().__init__(**kwargs)

    def share_module(self, module):
        self.service.share_module(module)

    def remove_shared_module(self, module):
        self.service.remove_shared_module(module)


class QudiNamespaceServer(BaseServer):
    """Contains a RPyC server that serves all activated qudi modules as well as a reference to the
    running qudi instance locally without encryption.
    You can specify the port but the host will always be "localhost"/127.0.0.1
    See qudi.core.remotemodules.RemoteModuleServer if you want to expose qudi modules to non-local
    clients.
    Actual rpyc server runs in a QThread.
    """

    def __init__(self, qudi, name, port, force_remote_calls_by_value=False, parent=None):
        """
        Parameters
        ----------
        qudi : qudi.Qudi
            The governing qudi main application instance.
        name : str
            Server name (used as name for the associated QThread).
        port : int
            Port number the RPyC server should listen to.
        parent : PySide2.QtCore.QObject, optional
            Parent Qt QObject.

        Returns
        -------
        None
        """
        service_instance = QudiNamespaceService(
            qudi=qudi,
            force_remote_calls_by_value=force_remote_calls_by_value
        )
        super().__init__(parent=parent,
                         qudi=qudi,
                         service_instance=service_instance,
                         name=name,
                         host='localhost',
                         port=port)



# -*- coding: utf-8 -*-
"""
StatusVar object for qudi modules to allow storing of application status variables on disk.
These variables get stored during deactivation of qudi modules and loaded back in during activation.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['StatusVar']

import copy
import inspect
from typing import Callable, Any, Optional


class StatusVar:
    """This class defines a status variable that is loaded before activation and saved after
    deactivation.
    """

    def __init__(self, name: Optional[str] = None, default: Optional[Any] = None, *,
                 constructor: Optional[Callable] = None, representer: Optional[Callable] = None):
        """
        Parameters
        ----------
        name : str
            Identifier of the status variable when stored.
        default : Any
            Default value for the status variable when a saved version is not present.
        constructor : callable
            Constructor function for the variable; use for type checks or conversion.
        representer : callable
            Representer function for the status variable; use for saving conversion.
        """
        self.name = name
        self.default = default
        self.constructor_function = None
        self.representer_function = None
        if constructor is not None:
            self.constructor(constructor)
        if representer is not None:
            self.representer(representer)

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict={}):
        return self.copy()

    def copy(self, **kwargs):
        """Create a new instance of StatusVar with copied and updated values.

        Parameters
        ----------
        **kwargs : dict
            Additional or overridden parameters for the constructor of this class.
        """
        newargs = {'name': self.name,
                   'default': copy.deepcopy(self.default),
                   'constructor': self.constructor_function,
                   'representer': self.representer_function}
        newargs.update(kwargs)
        return StatusVar(**newargs)

    def constructor(self, func: Callable) -> Callable:
        """This is the decorator for declaring constructor function for this StatusVar.

        Parameters
        ----------
        func : callable
            Constructor function for this StatusVar.

        Returns
        -------
        callable
            The original function so this can be used as a decorator.
        """
        self.constructor_function = self._assert_func_signature(func)
        return func

    def representer(self, func: Callable) -> Callable:
        """This is the decorator for declaring a representer function for this StatusVar.

        Parameters
        ----------
        func : callable
            Representer function for this StatusVar.

        Returns
        -------
        callable
            The original function so this can be used as a decorator.
        """
        self.representer_function = self._assert_func_signature(func)
        return func

    @staticmethod
    def _assert_func_signature(func: Callable) -> Callable:
        assert callable(func), 'StatusVar constructor/representer must be callable'
        params = tuple(inspect.signature(func).parameters)
        assert 0 < len(params) < 3, 'StatusVar constructor/representer must be function with ' \
                                    '1 (static) or 2 (bound method) parameters.'
        if len(params) == 1:
            def wrapper(instance, value):
                return func(value)

            return wrapper
        return func



# -*- coding: utf-8 -*-

"""
Improved pyqtgraph.AxisItem objects.

Copyright (c) 2022, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['LabelNudgeAxis', 'label_nudged_plot_widget']

from PySide2.QtCore import QSizeF, QPointF
from pyqtgraph import AxisItem, PlotWidget
from typing import Type


def label_nudged_plot_widget(plot_widget_type: Type[PlotWidget]) -> Type[PlotWidget]:
    class NudgedPlotWidget(plot_widget_type):
        def __init__(self, **kwargs) -> None:
            if 'axisItems' not in kwargs:
                bottom_axis = LabelNudgeAxis(orientation='bottom')
                left_axis = LabelNudgeAxis(orientation='left')
                bottom_axis.nudge = 0
                left_axis.nudge = 0
                kwargs['axisItems'] = {'bottom': bottom_axis, 'left': left_axis}
            super().__init__(**kwargs)
    return NudgedPlotWidget


class LabelNudgeAxis(AxisItem):
    """ This is a custom axis that extends the normal pyqtgraph to be able to nudge the axis labels
    """

    @property
    def nudge(self):
        if not hasattr(self, '_nudge'):
            self._nudge = 5
        return self._nudge

    @nudge.setter
    def nudge(self, nudge):
        self._nudge = nudge
        s = self.size()
        # call resizeEvent indirectly
        self.resize(s + QSizeF(1, 1))
        self.resize(s)

    def resizeEvent(self, ev=None):
        # Set the position of the label
        nudge = self.nudge
        br = self.label.boundingRect()
        p = QPointF(0, 0)
        size = self.size()
        if self.orientation == 'left':
            p.setY(int(size.height() / 2 + br.width() / 2))
            p.setX(-nudge)
        elif self.orientation == 'right':
            p.setY(int(size.height() / 2 + br.width() / 2))
            p.setX(int(size.width() - br.height() + nudge))
        elif self.orientation == 'top':
            p.setY(-nudge)
            p.setX(int(size.width() / 2.0 - br.width() / 2.0))
        elif self.orientation == 'bottom':
            p.setX(int(size.width() / 2.0 - br.width() / 2.0))
            p.setY(int(size.height() - br.height() + nudge))
        self.label.setPos(p)
        self.picture = None
